# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

"""
Mounts NCCH containers, creating a virtual filesystem of decrypted sections.
"""

import logging
import os
from errno import ENOENT
from math import ceil
from stat import S_IFDIR, S_IFREG
from sys import argv
from threading import Thread
from typing import BinaryIO, TYPE_CHECKING

from pyctr.crypto import CryptoEngine, Keyslot
from pyctr.types.ncch import NCCHReader, FIXED_SYSTEM_KEY
from pyctr.util import readbe, roundup
from . import _common as _c
# _common imports these from fusepy, and prints an error if it fails; this allows less duplicated code
from ._common import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context
from .exefs import ExeFSMount
from .romfs import RomFSMount

if TYPE_CHECKING:
    from typing import Dict, List


class NCCHContainerMount(LoggingMixIn, Operations):
    fd = 0
    romfs_fuse = None
    exefs_fuse = None

    def __init__(self, ncch_fp: BinaryIO, g_stat: os.stat_result, decompress_code: bool = True, dev: bool = False,
                 seeddb: str = None):
        self.crypto = CryptoEngine(dev=dev)

        self.decompress_code = decompress_code
        self.seeddb = seeddb
        self.files: Dict[str, Dict] = {}

        # get status change, modify, and file access times
        self._g_stat = g_stat
        self.g_stat = {'st_ctime': int(g_stat.st_ctime), 'st_mtime': int(g_stat.st_mtime),
                       'st_atime': int(g_stat.st_atime)}

        ncch_header = ncch_fp.read(0x200)
        self.reader = NCCHReader.from_header(ncch_header)

        self.f = ncch_fp

        if not self.reader.flags.no_crypto:
            # I should figure out what happens if fixed-key crypto is
            #   used along with seed. even though this will never
            #   happen in practice, I would still like to see what
            #   happens if it happens.
            if self.reader.flags.fixed_crypto_key:
                normal_key = FIXED_SYSTEM_KEY if self.reader.program_id & (0x10 << 32) else 0x0
                self.crypto.set_normal_key(Keyslot.NCCH, normal_key.to_bytes(0x10, 'big'))
            else:
                if self.reader.flags.uses_seed:
                    self.reader.load_seed_from_seeddb()

                self.crypto.set_keyslot('y', Keyslot.NCCH, readbe(self.reader.get_key_y(original=True)))
                self.crypto.set_keyslot('y', self.reader.extra_keyslot,
                                        readbe(self.reader.get_key_y()))

    def __del__(self, *args):
        try:
            self.f.close()
        except AttributeError:
            pass

    destroy = __del__

    def init(self, path, _setup_romfs=True):
        decrypted_filename = '/decrypted.' + ('cxi' if self.reader.flags.executable else 'cfa')

        self.files[decrypted_filename] = {'size': self.reader.content_size, 'offset': 0, 'enctype': 'fulldec'}
        self.files['/ncch.bin'] = {'size': 0x200, 'offset': 0, 'enctype': 'none'}

        if self.reader.check_for_extheader():
            self.files['/extheader.bin'] = {'size': 0x800, 'offset': 0x200, 'enctype': 'normal',
                                            'keyslot': Keyslot.NCCH,
                                            'iv': (self.reader.partition_id << 64 | (0x01 << 56))}

        plain_region = self.reader.plain_region
        if plain_region.offset:
            self.files['/plain.bin'] = {'size': plain_region.size, 'offset': plain_region.offset, 'enctype': 'none'}

        logo_region = self.reader.logo_region
        if logo_region.offset:
            self.files['/logo.bin'] = {'size': logo_region.size, 'offset': logo_region.offset, 'enctype': 'none'}

        exefs_region = self.reader.exefs_region
        if exefs_region.offset:
            exefs_type = 'exefs'
            if self.reader.extra_keyslot == Keyslot.NCCH:
                exefs_type = 'normal'
            self.files['/exefs.bin'] = {'size': exefs_region.size, 'offset': exefs_region.offset, 'enctype': exefs_type,
                                        'keyslot': Keyslot.NCCH, 'keyslot_extra': self.reader.extra_keyslot,
                                        'iv': (self.reader.partition_id << 64 | (0x02 << 56)),
                                        'keyslot_normal_range': [(0, 0x200)]}

            # noinspection PyBroadException
            try:
                # get code compression bit
                decompress = False
                if self.decompress_code and self.reader.check_for_extheader():
                    exh_flag = self.read('/extheader.bin', 1, 0xD, 0)
                    decompress = exh_flag[0] & 1
                exefs_vfp = _c.VirtualFileWrapper(self, '/exefs.bin', exefs_region.size)
                exefs_fuse = ExeFSMount(exefs_vfp, self._g_stat, decompress_code=decompress, strict=True)
                self.exefs_fuse = exefs_fuse
            except Exception as e:
                print(f'Failed to mount ExeFS: {type(e).__name__}: {e}')
                self.exefs_fuse = None
            else:
                if not self.reader.flags.no_crypto:
                    for n, ent in self.exefs_fuse.reader.entries.items():
                        if n in {'icon', 'banner'}:
                            self.files['/exefs.bin']['keyslot_normal_range'].append(
                                (ent.offset + 0x200, ent.offset + 0x200 + roundup(ent.size, 0x200)))

        if not self.reader.flags.no_romfs:
            romfs_region = self.reader.romfs_region
            if romfs_region.offset:
                self.files['/romfs.bin'] = {'size': romfs_region.size, 'offset': romfs_region.offset,
                                            'enctype': 'normal', 'keyslot': self.reader.extra_keyslot,
                                            'iv': (self.reader.partition_id << 64 | (0x03 << 56))}

        if _setup_romfs:
            self.setup_romfs()

        if self.exefs_fuse and '/code.bin' in self.exefs_fuse.files:
            if self.exefs_fuse.decompress_code:
                # the data is read here to avoid an issue with threading
                # (yes i am kind of lazy)
                print('ExeFS: Reading .code...')
                data = self.exefs_fuse.read('/code.bin', self.exefs_fuse.files['/code.bin'].size, 0, 0)
            else:
                data = None
            Thread(target=self.exefs_fuse.init, daemon=True, args=(path, data)).start()

    def setup_romfs(self):
        if '/romfs.bin' in self.files:
            # noinspection PyBroadException
            try:
                romfs_vfp = _c.VirtualFileWrapper(self, '/romfs.bin', self.reader.romfs_region.size)
                # noinspection PyTypeChecker
                romfs_fuse = RomFSMount(romfs_vfp, self._g_stat)
                romfs_fuse.init('/')
                self.romfs_fuse = romfs_fuse
            except Exception as e:
                print(f'Failed to mount RomFS: {type(e).__name__}: {e}')

    def flush(self, path, fh):
        return self.f.flush()

    @_c.ensure_lower_path
    def getattr(self, path, fh=None):
        if path.startswith('/exefs/'):
            return self.exefs_fuse.getattr(_c.remove_first_dir(path), fh)
        elif path.startswith('/romfs/'):
            return self.romfs_fuse.getattr(_c.remove_first_dir(path), fh)
        uid, gid, pid = fuse_get_context()
        if path in {'/', '/romfs', '/exefs'}:
            st = {'st_mode': (S_IFDIR | 0o555), 'st_nlink': 2}
        elif path in self.files:
            st = {'st_mode': (S_IFREG | 0o444), 'st_size': self.files[path]['size'], 'st_nlink': 1}
        else:
            raise FuseOSError(ENOENT)
        return {**st, **self.g_stat, 'st_uid': uid, 'st_gid': gid}

    def open(self, path, flags):
        self.fd += 1
        return self.fd

    @_c.ensure_lower_path
    def readdir(self, path, fh):
        if path.startswith('/exefs'):
            yield from self.exefs_fuse.readdir(_c.remove_first_dir(path), fh)
        elif path.startswith('/romfs'):
            yield from self.romfs_fuse.readdir(_c.remove_first_dir(path), fh)
        elif path == '/':
            yield from ('.', '..')
            yield from (x[1:] for x in self.files)
            if self.exefs_fuse is not None:
                yield 'exefs'
            if self.romfs_fuse is not None:
                yield 'romfs'

    @_c.ensure_lower_path
    def read(self, path, size, offset, fh):
        if path.startswith('/exefs/'):
            return self.exefs_fuse.read(_c.remove_first_dir(path), size, offset, fh)
        elif path.startswith('/romfs/'):
            return self.romfs_fuse.read(_c.remove_first_dir(path), size, offset, fh)
        fi = self.files[path]
        real_offset = fi['offset'] + offset
        if fi['offset'] + offset > fi['offset'] + fi['size']:
            return b''
        if offset + size > fi['size']:
            size = fi['size'] - offset

        if fi['enctype'] == 'none' or self.reader.flags.no_crypto:
            # if no encryption, just read and return
            self.f.seek(real_offset)
            data = self.f.read(size)

        elif fi['enctype'] == 'normal':
            self.f.seek(real_offset)
            data = self.f.read(size)
            # thanks Stary2001
            before = offset % 16
            after = (offset + size) % 16
            data = (b'\0' * before) + data + (b'\0' * after)
            iv = fi['iv'] + (offset >> 4)
            data = self.crypto.create_ctr_cipher(fi['keyslot'], iv).decrypt(data)[before:size + before]

        elif fi['enctype'] == 'exefs':
            # thanks Stary2001
            before = offset % 0x200
            aligned_real_offset = real_offset - before
            aligned_offset = offset - before
            aligned_size = size + before
            self.f.seek(aligned_real_offset)

            def do_thing(al_offset: int, al_size: int, cut_start: int, cut_end: int):
                end: int = al_offset + (ceil(al_size / 0x200) * 0x200)
                last_chunk_offset = end - 0x200
                # noinspection PyTypeChecker
                for chunk in range(al_offset, end, 0x200):
                    iv = fi['iv'] + (chunk >> 4)
                    keyslot = fi['keyslot_extra']
                    for r in fi['keyslot_normal_range']:
                        if r[0] <= self.f.tell() - fi['offset'] < r[1]:
                            keyslot = fi['keyslot']
                    out = self.crypto.create_ctr_cipher(keyslot, iv).decrypt(self.f.read(0x200))
                    if chunk == al_offset:
                        out = out[cut_start:]
                    if chunk == last_chunk_offset and cut_end != 0x200:
                        out = out[:-cut_end]
                    yield out

            data = b''.join(do_thing(aligned_offset, aligned_size, before, 0x200 - ((size + before) % 0x200)))

        elif fi['enctype'] == 'fulldec':
            # this could be optimized much better
            before = offset % 0x200
            aligned_real_offset = real_offset - before
            aligned_offset = offset - before
            aligned_size = size + before
            self.f.seek(aligned_real_offset)

            def do_thing(al_offset: int, al_size: int, cut_start: int, cut_end: int):
                end: int = al_offset + (ceil(al_size / 0x200) * 0x200)
                # dict is ordered by default in CPython since 3.6.0
                # and part of the language spec since 3.7.0
                to_read: Dict[str, List[int]] = {}

                if self.reader.check_for_extheader():
                    extheader_start = 0x200
                    extheader_end = 0xA00
                else:
                    extheader_start = extheader_end = 0

                logo = self.reader.logo_region
                logo_start = logo.offset
                logo_end = logo_start + logo.size

                plain = self.reader.plain_region
                plain_start = plain.offset
                plain_end = plain_start + plain.size

                exefs = self.reader.exefs_region
                exefs_start = exefs.offset
                exefs_end = exefs_start + exefs.size

                romfs = self.reader.romfs_region
                romfs_start = romfs.offset
                romfs_end = romfs_start + romfs.size

                for chunk_offset in range(al_offset, end, 0x200):
                    # RomFS check first, since it might be faster
                    if romfs_start <= chunk_offset < romfs_end:
                        name = '/romfs.bin'
                        curr_offset = romfs_start
                    # ExeFS check second, since it might be faster
                    elif exefs_start <= chunk_offset < exefs_end:
                        name = '/exefs.bin'
                        curr_offset = exefs_start
                    # NCCH check, always 0x0 to 0x200
                    elif 0 <= chunk_offset < 0x200:
                        name = '/ncch.bin'
                        curr_offset = 0
                    elif extheader_start <= chunk_offset < extheader_end:
                        name = '/extheader.bin'
                        curr_offset = extheader_start
                    elif logo_start <= chunk_offset < logo_end:
                        name = '/logo.bin'
                        curr_offset = logo_start
                    elif plain_start <= chunk_offset < plain_end:
                        name = '/plain.bin'
                        curr_offset = plain_start
                    else:
                        name = f'raw{chunk_offset}'
                        curr_offset = 0
                    if name not in to_read:
                        to_read[name] = [chunk_offset - curr_offset, 0]
                    to_read[name][1] += 0x200
                    last_name = name

                is_start = True
                for name, info in to_read.items():
                    try:
                        new_data = self.read(name, info[1], info[0], 0)
                        if name == '/ncch.bin':
                            # fix crypto flags
                            ncch_array = bytearray(new_data)
                            ncch_array[0x18B] = 0
                            ncch_array[0x18F] = 4
                            new_data = bytes(ncch_array)
                    except KeyError:
                        # for unknown files
                        self.f.seek(info[0])
                        new_data = self.f.read(info[1])
                    if is_start is True:
                        new_data = new_data[cut_start:]
                        is_start = False
                    # noinspection PyUnboundLocalVariable
                    if name == last_name and cut_end != 0x200:
                        new_data = new_data[:-cut_end]

                    yield new_data

            data = b''.join(do_thing(aligned_offset, aligned_size, before, 0x200 - ((size + before) % 0x200)))

        else:
            from pprint import pformat
            print('--------------------------------------------------',
                  'Warning: unknown file type (this should not happen!)',
                  'Please file an issue or contact the developer with the details below.',
                  '  https://github.com/ihaveamac/ninfs/issues',
                  '--------------------------------------------------',
                  f'{path!r}: {pformat(fi)!r}', sep='\n')

            data = b'g' * size

        return data

    @_c.ensure_lower_path
    def statfs(self, path):
        if path.startswith('/exefs/'):
            return self.exefs_fuse.statfs(_c.remove_first_dir(path))
        elif path.startswith('/romfs/'):
            return self.romfs_fuse.statfs(_c.remove_first_dir(path))
        else:
            return {'f_bsize': 4096, 'f_blocks': self.reader.content_size // 4096, 'f_bavail': 0, 'f_bfree': 0,
                    'f_files': len(self.files)}


def main(prog: str = None, args: list = None):
    from argparse import ArgumentParser
    if args is None:
        args = argv[1:]
    parser = ArgumentParser(prog=prog, description='Mount Nintendo 3DS NCCH containers.',
                            parents=(_c.default_argp, _c.dev_argp, _c.seeddb_argp,
                                     _c.main_args('ncch', 'NCCH file')))

    a = parser.parse_args(args)
    opts = dict(_c.parse_fuse_opts(a.o))

    if a.do:
        logging.basicConfig(level=logging.DEBUG, filename=a.do)

    ncch_stat = os.stat(a.ncch)

    with open(a.ncch, 'rb') as f:
        mount = NCCHContainerMount(ncch_fp=f, dev=a.dev, g_stat=ncch_stat, seeddb=a.seeddb)
        if _c.macos or _c.windows:
            opts['fstypename'] = 'NCCH'
            if _c.macos:
                opts['volname'] = f'NCCH Container ({mount.reader.product_code}; {mount.reader.program_id:016X})'
            elif _c.windows:
                # volume label can only be up to 32 chars
                opts['volname'] = f'NCCH ({mount.reader.product_code})'
        FUSE(mount, a.mount_point, foreground=a.fg or a.do or a.d, ro=True, nothreads=True, debug=a.d,
             fsname=os.path.realpath(a.ncch).replace(',', '_'), **opts)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

# not very good with gui development...
# don't read this file, it sucks
print('Importing dependencies...')

import json
import webbrowser
from configparser import ConfigParser
from contextlib import suppress
from glob import iglob
from hashlib import sha256
from os import environ, kill, makedirs, rmdir
from os.path import abspath, dirname, expanduser, getsize, isfile, isdir, ismount, join as pjoin
from shutil import get_terminal_size
from ssl import PROTOCOL_TLSv1_2, SSLContext
from subprocess import Popen, check_call, CalledProcessError
from sys import argv, executable, exit, platform, version_info, maxsize, stderr
from time import sleep
from traceback import print_exc
from typing import TYPE_CHECKING
from urllib.request import urlopen

try:
    from appJar import gui
    from appJar.appjar import ItemLookupError
except ImportError as e:
    print_exc()
    print(file=stderr)
    if 'tk' in e.args[0].lower():
        exit('Could not import tkinter, please install python3-tk (or equivalent for your distribution).')
    else:
        exit('Could not import appJar. If you installed via pip, make sure you installed with the gui.\n'
             'Read the README at the GitHub repository.')
    # this code will never be reached
    gui = ItemLookupError = None

from pkg_resources import parse_version

from __init__ import __version__ as version
from fmt_detect import detect_format
from pyctr.util import config_dirs

if TYPE_CHECKING:
    from http.client import HTTPResponse
    # noinspection PyProtectedMember
    from pkg_resources._vendor.packaging.version import Version
    from typing import Any, Dict, List

windows = platform == 'win32'  # only for native windows, not cygwin
macos = platform == 'darwin'

ITEM = 'item'
EASTWEST = 'ew'
PV = 'previous'
LABEL1 = 'label1'
LABEL2 = 'label2'
LABEL3 = 'label3'
MOUNTPOINT = 'mountpoint'

MOUNT = 'Mount'
UNMOUNT = 'Unmount'
DIRECTORY = 'Directory'
FILE = 'File'
OK = 'OK'

DRAGFILE = 'Drag a file here or browse...'
DRAGDIR = 'Drag a directory here or browse...'
BROWSE = 'Browse...'
ALLOW_OTHER = 'Allow access by other users (-o allow_other)'
ALLOW_ROOT = 'Allow access by root (-o allow_root)'
ALLOW_NONE = "Don't allow other users"

b9_hashes = {
    '2f88744feed717856386400a44bba4b9ca62e76a32c715d4f309c399bf28166f': 'boot9',
    '7331f7edece3dd33f2ab4bd0b3a5d607229fd19212c10b734cedcaf78c1a7b98': 'boot9_prot',
}

b9_paths: 'List[str]' = []
for p in config_dirs:
    b9_paths.append(pjoin(p, 'boot9.bin'))
    b9_paths.append(pjoin(p, 'boot9_prot.bin'))

with suppress(KeyError):
    b9_paths.insert(0, environ['BOOT9_PATH'])

seeddb_paths: 'List[str]' = [pjoin(x, 'seeddb.bin') for x in config_dirs]
with suppress(KeyError):
    seeddb_paths.insert(0, environ['SEEDDB_PATH'])

home = expanduser('~')

# dir to copy to if chosen to copy boot9/seeddb in gui
# config_dir is fuse-3ds because it is the previous name of this project, maybe later I'll support a new ninfs dir
if windows:
    target_dir: str = pjoin(environ.get('APPDATA'), '3ds')
    config_dir: str = pjoin(environ.get('APPDATA'), 'fuse-3ds')
elif macos:
    target_dir: str = pjoin(home, 'Library', 'Application Support', '3ds')
    config_dir: str = pjoin(home, 'Library', 'Application Support', 'fuse-3ds')
else:
    target_dir: str = pjoin(home, '.3ds')
    config_root = environ.get('XDG_CONFIG_HOME')
    if not config_root:
        # check other paths in XDG_CONFIG_DIRS to see if fuse-3ds already exists in one of them
        config_roots: str = environ.get('XDG_CONFIG_DIRS')
        if not config_roots:
            config_roots = '/etc/xdg'
        config_paths = config_roots.split(':')
        for p in config_paths:
            d = pjoin(p, 'fuse-3ds')
            if isdir(d):
                config_root = p
                break
    # check again to see if it was set
    if not config_root:
        config_root = pjoin(home, '.config')
    config_dir = pjoin(config_root, 'fuse-3ds')

makedirs(config_dir, exist_ok=True)
update_config = ConfigParser()

configs = {'update': update_config}


def init_config(kind: str, defaults):
    config_path = pjoin(config_dir, kind + '.cfg')
    if not configs[kind].read(config_path):
        print('Creating new', kind, 'config...')
        configs[kind].update(defaults)
        write_config(kind)
    else:
        print('Loaded', kind, 'config.')


def write_config(kind: str, section=None, option=None, value=None):
    config_path = pjoin(config_dir, kind + '.cfg')
    if section:
        configs[kind][section][option] = str(value)
    try:
        with open(config_path, 'w', encoding='utf-8') as o:
            configs[kind].write(o)
            print('Wrote', kind, 'config to', config_path)
            return True
    except Exception:
        print_exc()
        print()
        print('Failed to write', kind, 'config to', config_path)
        return False


init_config('update', {'update': {'check_updates_online': True, 'ignored_update': 'v0.0'}})


for p in b9_paths:
    if isfile(p):
        b9_found = p
        break
else:
    b9_found = ''

for p in seeddb_paths:
    if isfile(p):
        seeddb_found = p
        break
else:
    seeddb_found = ''

# types
CCI = 'CTR Cart Image (".3ds", ".cci")'
CDN = 'CDN contents ("cetk", "tmd", and contents)'
CIA = 'CTR Importable Archive (".cia")'
EXEFS = 'Executable Filesystem (".exefs", "exefs.bin")'
NANDCTR = 'Nintendo 3DS NAND backup ("nand.bin")'
NANDTWL = 'Nintendo DSi NAND backup ("nand_dsi.bin")'
NANDHAC = 'Nintendo Switch NAND backup ("rawnand.bin")'
NCCH = 'NCCH (".cxi", ".cfa", ".ncch", ".app")'
ROMFS = 'Read-only Filesystem (".romfs", "romfs.bin")'
SD = 'SD Card Contents ("Nintendo 3DS" from SD)'
SRL = 'Nintendo DS ROM image (".nds", ".srl")'
THREEDSX = '3DSX Homebrew (".3dsx")'
TITLEDIR = 'Titles directory ("title" from NAND or SD)'

mount_types = {CCI: 'cci', CDN: 'cdn', CIA: 'cia', EXEFS: 'exefs', NANDCTR: 'nandctr', NANDHAC: 'nandhac',
               NANDTWL: 'nandtwl', NCCH: 'ncch', ROMFS: 'romfs', SD: 'sd', SRL: 'srl', THREEDSX: 'threedsx',
               TITLEDIR: 'titledir'}

mount_types_rv: 'Dict[str, str]' = {y: x for x, y in mount_types.items()}

ctr_types = (CCI, CDN, CIA, EXEFS, NANDCTR, NCCH, ROMFS, SD, THREEDSX, TITLEDIR)
twl_types = (NANDTWL, SRL)
hac_types = (NANDHAC,)
types_list = ctr_types + twl_types

types_requiring_b9 = {CCI, CDN, CIA, NANDCTR, NCCH, SD, TITLEDIR}

if windows:
    from ctypes import windll
    from platform import win32_ver
    from signal import CTRL_BREAK_EVENT
    from string import ascii_uppercase
    from subprocess import CREATE_NEW_PROCESS_GROUP
    from sys import stdout
    # noinspection PyUnresolvedReferences
    from winreg import OpenKey, QueryValueEx, HKEY_LOCAL_MACHINE

    from reg_shell import add_reg, del_reg, uac_enabled

    # unlikely, but this causes issues
    if stdout is None:  # happens if pythonw is used on windows
        # this one uses MessageBoxW directly because the gui hasn't been made yet.
        res = windll.user32.MessageBoxW(None, (
            'This is being run with the wrong Python executable.\n'
            'This should be installed as a module, then run using the py launcher on Python 3.6.1 or later.\n\n'
            'Click OK to open the ninfs repository on GitHub:\n'
            'https://github.com/ihaveamac/ninfs'),
                                        'ninfs', 0x00000010 | 0x00000001)
        if res == 1:
            webbrowser.open('https://github.com/ihaveamac/ninfs')
        exit(1)


    def get_unused_drives():
        # https://stackoverflow.com/questions/827371/is-there-a-way-to-list-all-the-available-drive-letters-in-python
        drives: List[str] = []
        bitmask = windll.kernel32.GetLogicalDrives()
        for letter in ascii_uppercase:
            if not bitmask & 1:
                drives.append(letter)
            bitmask >>= 1

        return drives


    def update_drives():
        o = get_unused_drives()
        app.changeOptionBox(MOUNTPOINT, (x + ':' for x in o))
        app.setOptionBox(MOUNTPOINT, o[-1] + ':')
elif macos:
    from platform import mac_ver

_used_pyinstaller = False
_ndw_resp = False
process: Popen = None
curr_mountpoint: str = None

pyver = f'{version_info[0]}.{version_info[1]}.{version_info[2]}'
if version_info[3] != 'final':
    pyver += f'{version_info[3][0]}{version_info[4]}'
pybits = 64 if maxsize > 0xFFFFFFFF else 32

app = gui('ninfs v' + version, showIcon=False, handleArgs=False)
if windows:
    app.setIcon(pjoin(dirname(__file__), 'data', 'windows.ico'))


def run_mount(module_type: str, item: str, mountpoint: str, extra_args: list = ()):
    global process, curr_mountpoint
    if process is None or process.poll() is not None:
        environ['BOOT9_PATH'] = b9_found
        environ['SEEDDB_PATH'] = seeddb_found
        args = [executable]
        if not _used_pyinstaller:
            args.append(dirname(__file__))
        args.extend((module_type, '-f', item, mountpoint))
        args.extend(extra_args)
        curr_mountpoint = mountpoint
        x, _ = get_terminal_size()
        print('-' * (x - 1))
        print('Running:', args)
        opts = {}
        if windows:
            opts['creationflags'] = CREATE_NEW_PROCESS_GROUP
        process = Popen(args, **opts)

        # check if the mount exists, or if the process exited before it
        check = isdir if windows else ismount
        while not check(mountpoint):
            sleep(1)
            if process.poll() is not None:
                show_mounterror()
                app.queueFunction(app.enableButton, MOUNT)
                return

        app.queueFunction(app.enableButton, UNMOUNT)

        if windows:
            with suppress(CalledProcessError):
                # not using startfile since i've been getting fatal errors (PyEval_RestoreThread) on windows
                #   for some reason
                # also this error always appears when calling explorer, so i'm ignoring it
                check_call(['explorer', mountpoint.replace('/', '\\')])
        elif macos:
            try:
                check_call(['/usr/bin/open', '-a', 'Finder', mountpoint])
            except CalledProcessError:
                print(f'Failed to open Finder on {mountpoint}')
                print_exc()
        else:  # probably linux
            try:
                check_call(['xdg-open', mountpoint])
            except CalledProcessError:
                print(f'Failed to open the file manager on {mountpoint}')
                print_exc()

        if process.wait() != 0:
            # just in case there are leftover mounts
            try:
                stop_mount()
            except CalledProcessError:
                print_exc()
            show_exiterror(process.returncode)

        app.queueFunction(app.disableButton, UNMOUNT)
        app.queueFunction(app.enableButton, MOUNT)


def stop_mount():
    if process is not None and process.poll() is None:
        print('Stopping')
        if windows:
            kill(process.pid, CTRL_BREAK_EVENT)
        else:
            # this is cheating...
            if macos:
                check_call(['diskutil', 'unmount', curr_mountpoint])
            else:
                # assuming linux or bsd, which have fusermount
                check_call(['fusermount', '-u', curr_mountpoint])


def press(button: str):
    if button == MOUNT:
        extra_args = []
        mount_type = app.getOptionBox('TYPE')
        app.disableButton(MOUNT)
        item = app.getEntry(mount_type + ITEM)
        if not item:
            show_noitemerror()
            app.enableButton(MOUNT)
            return

        if windows:
            if app.getRadioButton('mountpoint-choice') == 'Drive letter':
                if mount_type in {NANDCTR, NANDHAC, NANDTWL}:
                    res = app.okBox(
                        'ninfs Warning',
                        'You chose drive letter when using the NAND mount.\n'
                        '\n'
                        'Using a directory mount over a drive letter for NAND is highly '
                        'recommended because some tools like OSFMount will not be '
                        'able to read from files in a mount using a drive letter.\n'
                        '\n'
                        'Are you sure you want to continue?'
                    )
                    if not res:
                        app.enableButton(MOUNT)
                        return
                mountpoint = app.getOptionBox(MOUNTPOINT)
            else:
                mountpoint = app.getEntry(MOUNTPOINT)
                if not mountpoint:
                    show_nomperror()
                    app.enableButton(MOUNT)
                    return
                try:
                    # winfsp won't work on an existing directory
                    # so we try to use rmdir, which will delete it, only if it's empty
                    rmdir(mountpoint)
                except FileNotFoundError:
                    pass
                except Exception as e:
                    # noinspection PyUnresolvedReferences
                    if isinstance(e, OSError) and e.winerror == 145:  # "The directory is not empty"
                        show_mounterror_dir_win()
                    else:
                        print_exc()
                        show_mounterror()
                    app.enableButton(MOUNT)
                    return
        else:
            mountpoint = app.getEntry(MOUNTPOINT)
            if not mountpoint:
                show_nomperror()
                app.enableButton(MOUNT)
                return

        if mount_type == CDN:
            key = app.getEntry(CDN + 'key')
            if key:
                try:
                    _res = bytes.fromhex(key)
                except ValueError:
                    app.warningBox('ninfs Error', 'The given titlekey was not a valid hexstring.')
                    app.enableButton(MOUNT)
                    return
                if len(_res) != 16:
                    app.warningBox('ninfs Error', 'The given titlekey must be 32 characters.')
                    app.enableButton(MOUNT)
                    return
                extra_args.extend(('--dec-key', key))
        elif mount_type == NANDCTR:
            otp = app.getEntry(NANDCTR + 'otp')
            aw = app.getCheckBox(NANDCTR + 'aw')
            if otp:
                extra_args.extend(('--otp', otp))
            if not aw:
                extra_args.append('-r')
        elif mount_type == NANDHAC:
            try:
                int(item[-2:])
                # since int conversion succeded, assume this is a split file
                extra_args.append('--split-files')
            except ValueError:
                pass
            bis = app.getEntry(NANDHAC + 'bis')
            aw = app.getCheckBox(NANDHAC + 'aw')
            if not bis:
                app.warningBox('ninfs Error', 'BIS keys are required.')
                app.enableButton(MOUNT)
                return
            extra_args.extend(('--keys', bis))
            if not aw:
                extra_args.append('-r')
        elif mount_type == NANDTWL:
            consoleid = app.getEntry(NANDTWL + 'consoleid')
            aw = app.getCheckBox(NANDTWL + 'aw')
            if consoleid:
                if consoleid.lower().startswith('fw'):
                    app.warningBox('ninfs Error', 'A real Console ID does not start with FW. '
                                                     'It is a 16-character hexstring.')
                    app.enableButton(MOUNT)
                    return
                try:
                    _res = bytes.fromhex(consoleid)
                except ValueError:
                    app.warningBox('ninfs Error', 'The given Console ID was not a valid hexstring.')
                    app.enableButton(MOUNT)
                    return
                if len(_res) != 8:
                    app.warningBox('ninfs Error', 'The given Console ID must be 16 characters.')
                    app.enableButton(MOUNT)
                    return
                extra_args.extend(('--console-id', consoleid))
            if not aw:
                extra_args.append('-r')
        elif mount_type == SD:
            movable = app.getEntry(SD + 'movable')
            aw = app.getCheckBox(SD + 'aw')
            if not movable:
                app.warningBox('ninfs Error', 'A movable.sed is required.')
                app.enableButton(MOUNT)
                return
            extra_args.extend(('--movable', movable))
            if not aw:
                extra_args.append('-r')
        elif mount_type == TITLEDIR:
            decompress = app.getCheckBox(TITLEDIR + 'decompress')
            mount_all = app.getCheckBox(TITLEDIR + 'mountall')
            if decompress:
                extra_args.append('--decompress-code')
            if mount_all:
                extra_args.append('--mount-all')
        elif mount_type == EXEFS:
            decompress = app.getCheckBox(EXEFS + 'decompress')
            if decompress:
                extra_args.append('--decompress-code')

        if app.getCheckBox('debug'):
            extra_args.extend(('--do', app.getEntry('debug')))

        if not windows:
            allow_user = app.getRadioButton('allowuser')
            if allow_user == ALLOW_ROOT:
                extra_args.extend(('-o', 'allow_root'))
            elif allow_user == ALLOW_OTHER:
                extra_args.extend(('-o', 'allow_other'))

        if mount_type in types_requiring_b9:
            use_dev_keys = app.getCheckBox('devkeys')
            if use_dev_keys:
                extra_args.append('--dev')

        app.thread(run_mount, mount_types[mount_type], item, mountpoint, extra_args)

    elif button == UNMOUNT:
        app.disableButton(UNMOUNT)
        # noinspection PyBroadException
        try:
            stop_mount()
            app.enableButton(MOUNT)
        except Exception:
            print_exc()
            app.showSubWindow('unmounterror')
            app.enableButton(UNMOUNT)
    elif button == 'Settings & Help':
        show_extras()


def kill_process(_):
    process.kill()
    app.hideSubWindow('unmounterror')
    app.enableButton(MOUNT)
    app.disableButton(UNMOUNT)


current_type = 'default'


def check_mount_button(mount_type: str):
    if mount_type not in types_list:
        return
    if not b9_found and mount_type in types_requiring_b9:
        app.disableButton(MOUNT)
    else:
        if process is None or process.poll() is not None:
            app.enableButton(MOUNT)


def change_type(*_):
    global current_type
    mount_type: str = app.getOptionBox('TYPE')
    try:
        app.openLabelFrame('Mount settings')
    except ItemLookupError:
        app.setSticky(EASTWEST)
        app.startLabelFrame('Mount settings', row=1, colspan=3)
        app.setSticky(EASTWEST)
        app.showLabelFrame('Mount settings')

    app.hideFrame(current_type)

    try:
        app.showFrame(mount_type)
    except ItemLookupError:
        if mount_type == CCI:
            with app.frame(CCI, row=1, colspan=3):
                app.addLabel(CCI + LABEL1, FILE, row=0, column=0)
                app.addFileEntry(CCI + ITEM, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(CCI + ITEM, DRAGFILE)

        elif mount_type == CDN:
            with app.frame(CDN, row=1, colspan=3):
                app.addLabel(CDN + LABEL1, DIRECTORY, row=0, column=0)
                app.addDirectoryEntry(CDN + ITEM, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(CDN + ITEM, DRAGDIR)

                app.addLabel(CDN + LABEL2, 'Decrypted Titlekey*', row=3, column=0)
                app.addEntry(CDN + 'key', row=3, column=1, colspan=2)

                app.setEntryDefault(CDN + 'key', 'Insert a decrypted titlekey')
                app.addLabel(CDN + LABEL3, '*Not required if title has a cetk.', row=4, colspan=3)

        elif mount_type == CIA:
            with app.frame(CIA, row=1, colspan=3):
                app.addLabel(CIA + LABEL1, FILE, row=0, column=0)
                app.addFileEntry(CIA + ITEM, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(CIA + ITEM, DRAGFILE)

        elif mount_type == EXEFS:
            with app.frame(EXEFS, row=1, colspan=3):
                app.addLabel(EXEFS + LABEL1, FILE, row=0, column=0)
                app.addFileEntry(EXEFS + ITEM, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(EXEFS + ITEM, DRAGFILE)

                app.addLabel(EXEFS + LABEL3, 'Options', row=3, column=0)
                app.addNamedCheckBox('Decompress .code', EXEFS + 'decompress', row=3, column=1, colspan=1)

        elif mount_type == NANDCTR:
            with app.frame(NANDCTR, row=1, colspan=3):
                app.addLabel(NANDCTR + LABEL1, FILE, row=0, column=0)
                app.addFileEntry(NANDCTR + ITEM, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(NANDCTR + ITEM, DRAGFILE)

                app.addLabel(NANDCTR + LABEL2, 'OTP file*', row=2, column=0)
                app.addFileEntry(NANDCTR + 'otp', row=2, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(NANDCTR + 'otp', DRAGFILE)

                app.addLabel(NANDCTR + 'label4', '*Not required if backup has essential.exefs from GodMode9.', row=3,
                             colspan=3)

                app.addLabel(NANDCTR + 'label5', 'Options', row=4, column=0)
                app.addNamedCheckBox('Allow writing', NANDCTR + 'aw', row=4, column=1, colspan=1)

                if has_dnd:
                    app.setEntryDropTarget(NANDCTR + 'otp', make_dnd_entry_check(NANDCTR + 'otp'))

        elif mount_type == NANDHAC:
            with app.frame(NANDHAC, row=1, colspan=3):
                app.addLabel(NANDHAC + LABEL1, FILE, row=0, column=0)
                app.addFileEntry(NANDHAC + ITEM, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(NANDHAC + ITEM, DRAGFILE)

                app.addLabel(NANDHAC + LABEL2, 'BIS Keys', row=2, column=0)
                app.addFileEntry(NANDHAC + 'bis', row=2, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(NANDHAC + 'bis', DRAGFILE)

                app.addLabel(NANDCTR + LABEL3, 'Multi-part backups are supported (e.g. rawnand.bin.00)', row=3,
                             colspan=3)

                app.addLabel(NANDHAC + 'label4', 'Options', row=4, column=0)
                app.addNamedCheckBox('Allow writing', NANDHAC + 'aw', row=4, column=1, colspan=1)

                if has_dnd:
                    app.setEntryDropTarget(NANDHAC + 'bis', make_dnd_entry_check(NANDHAC + 'bis'))

        elif mount_type == NANDTWL:
            with app.frame(NANDTWL, row=1, colspan=3):
                app.addLabel(NANDTWL + LABEL1, FILE, row=0, column=0)
                app.addFileEntry(NANDTWL + ITEM, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(NANDTWL + ITEM, DRAGFILE)

                app.addLabel(NANDTWL + LABEL2, 'Console ID*', row=2, column=0)
                app.addEntry(NANDTWL + 'consoleid', row=2, column=1, colspan=2)
                app.setEntryDefault(NANDTWL + 'consoleid', 'If required, input Console ID as hexstring')

                app.addLabel(NANDTWL + LABEL3, '*Not required if backup has nocash footer with ConsoleID/CID.', row=3,
                             colspan=3)

                app.addLabel(NANDTWL + 'label4', 'Options', row=5, column=0)
                app.addNamedCheckBox('Allow writing', NANDTWL + 'aw', row=5, column=1, colspan=1)

        elif mount_type == NCCH:
            with app.frame(NCCH, row=1, colspan=3):
                app.addLabel(NCCH + LABEL1, FILE, row=0, column=0)
                app.addFileEntry(NCCH + ITEM, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(NCCH + ITEM, DRAGFILE)

        elif mount_type == ROMFS:
            with app.frame(ROMFS, row=1, colspan=3):
                app.addLabel(ROMFS + LABEL1, FILE, row=0, column=0)
                app.addFileEntry(ROMFS + ITEM, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(ROMFS + ITEM, DRAGFILE)

        elif mount_type == SD:
            with app.frame(SD, row=1, colspan=3):
                app.addLabel(SD + LABEL1, DIRECTORY, row=0, column=0)
                app.addDirectoryEntry(SD + ITEM, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(SD + ITEM, DRAGDIR)

                app.addLabel(SD + LABEL2, 'movable.sed', row=2, column=0)
                app.addFileEntry(SD + 'movable', row=2, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(SD + 'movable', DRAGFILE)

                app.addLabel(SD + LABEL3, 'Options', row=3, column=0)
                app.addNamedCheckBox('Allow writing', SD + 'aw', row=3, column=1, colspan=1)

                if has_dnd:
                    app.setEntryDropTarget(SD + 'movable', make_dnd_entry_check(SD + 'movable'))

        elif mount_type == THREEDSX:
            with app.frame(THREEDSX, row=1, colspan=3):
                app.addLabel(THREEDSX + LABEL1, FILE, row=0, column=0)
                app.addFileEntry(THREEDSX + ITEM, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(THREEDSX + ITEM, DRAGFILE)

        elif mount_type == SRL:
            with app.frame(SRL, row=1, colspan=3):
                app.addLabel(SRL + LABEL1, FILE, row=0, column=0)
                app.addFileEntry(SRL + ITEM, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(SRL + ITEM, DRAGFILE)

        elif mount_type == TITLEDIR:
            with app.frame(TITLEDIR, row=1, colspan=3):
                app.addLabel(TITLEDIR + LABEL1, DIRECTORY, row=0, column=0)
                app.addDirectoryEntry(TITLEDIR + ITEM, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                app.setEntryDefault(TITLEDIR + ITEM, DRAGFILE)

                app.addLabel(TITLEDIR + LABEL3, 'Options', row=3, column=0)
                app.addNamedCheckBox('Decompress .code (slow!)', TITLEDIR + 'decompress', row=3, column=1, colspan=1)
                app.addNamedCheckBox('Mount all contents', TITLEDIR + 'mountall', row=3, column=2, colspan=1)

        if has_dnd:
            app.setEntryDropTarget(mount_type + ITEM, make_dnd_entry_check(mount_type + ITEM))

    finally:
        app.stopLabelFrame()
        current_type = mount_type

        try:
            app.showLabelFrame('Mount point')
        except ItemLookupError:
            app.setSticky(EASTWEST)
            with app.labelFrame('Mount point', row=2, colspan=3):
                app.setSticky(EASTWEST)
                with app.frame('mountpoint-root', colspan=3):
                    if windows:
                        def rb_change(_):
                            if app.getRadioButton('mountpoint-choice') == 'Drive letter':
                                app.hideFrame('mountpoint-dir')
                                app.showFrame('mountpoint-drive')
                            else:
                                app.hideFrame('mountpoint-drive')
                                app.showFrame('mountpoint-dir')

                        app.addLabel('mountpoint-choice-label', 'Mount type', row=0)
                        app.addRadioButton('mountpoint-choice', 'Drive letter', row=0, column=1)
                        app.addRadioButton('mountpoint-choice', 'Directory', row=0, column=2)

                        app.setRadioButtonChangeFunction('mountpoint-choice', rb_change)
                        with app.frame('mountpoint-drive', row=1, colspan=3):
                            app.addLabel('mountlabel1', 'Drive letter', row=0, column=0)
                            # putting "WWWW" to avoid a warning
                            app.addOptionBox(MOUNTPOINT, ['WWWW'], row=0, column=1, colspan=2)

                        with app.frame('mountpoint-dir', row=1, colspan=3):
                            app.addLabel('mountlabel2', 'Mount point', row=0, column=0)
                            app.addDirectoryEntry(MOUNTPOINT, row=0, column=1, colspan=2).theButton.config(text=BROWSE)
                        app.hideFrame('mountpoint-dir')
                        # noinspection PyUnboundLocalVariable
                        update_drives()
                    else:
                        app.addLabel('mountlabel', 'Mount point', row=2, column=0)
                        app.addDirectoryEntry(MOUNTPOINT, row=2, column=1, colspan=2).theButton.config(text=BROWSE)
                    app.setEntryDefault(MOUNTPOINT, DRAGDIR)

                    def toggle_advanced_opts(_):
                        if app.getCheckBox('advopt'):
                            app.showLabelFrame('Advanced options')
                        else:
                            app.hideLabelFrame('Advanced options')

                    def choose_debug_location(_):
                        path: str = app.saveBox(fileName='ninfs.log', dirName=pjoin(home, 'Desktop'),
                                                fileTypes=(),
                                                fileExt='.log')
                        if path:
                            app.setEntry('debug', path)

                    app.addNamedCheckBox('Show advanced options', 'advopt', column=1, colspan=2)
                    app.setCheckBoxChangeFunction('advopt', toggle_advanced_opts)

                app.setSticky(EASTWEST)
                with app.labelFrame('Advanced options', colspan=3):
                    app.setSticky(EASTWEST)

                    app.addLabel('devkeys-label', 'Keys')
                    app.addNamedCheckBox('Use developer-unit keys', 'devkeys', row=PV, column=1, colspan=2)

                    if not windows:
                        app.addLabel('allowuser-label', 'External user access')
                        app.addRadioButton('allowuser', ALLOW_NONE, row=PV, column=1, colspan=2)
                        app.addRadioButton('allowuser', ALLOW_ROOT, column=1, colspan=2)
                        app.addRadioButton('allowuser', ALLOW_OTHER, column=1, colspan=2)
                        if not macos:
                            app.addMessage('allowuser-notice',
                                           'This may require extra permissions, such as adding the user to the '
                                           'fuse group, or editing /etc/fuse.conf.', column=1, colspan=2)
                            app.setMessageAlign('allowuser-notice', 'left')
                            app.setMessageAspect('allowuser-notice', 500)

                    app.addLabel('debug-label1', 'Debug output')
                    app.addNamedCheckBox('Enable debug output', 'debug', row=PV, column=1, colspan=2)
                    app.addLabel('debug-label2', 'Debug log file')
                    app.addNamedButton('Choose log location...', 'debug', choose_debug_location, row=PV, column=1,
                                       colspan=2)
                    app.addEntry('debug', column=1, colspan=2)
                    app.setEntryDefault('debug', 'Choose where to save above...')

                app.hideLabelFrame('Advanced options')

                app.addButtons([MOUNT, UNMOUNT], press, colspan=3)
                app.disableButton(UNMOUNT)

            if has_dnd:
                app.setEntryDropTarget(MOUNTPOINT, make_dnd_entry_check(MOUNTPOINT))

        check_mount_button(mount_type)

        try:
            if mount_type in types_requiring_b9:
                app.showFrame('FOOTER')
            else:
                app.hideFrame('FOOTER')
        except ItemLookupError:
            pass

        if mount_type in {NANDCTR, NANDHAC, NANDTWL} and windows:
            app.setRadioButton('mountpoint-choice', 'Directory')


def make_dnd_entry_check(entry_name: str):
    def handle(data: str):
        if data.startswith('{'):
            data = data[1:-1]
        app.setEntry(entry_name, data)

    return handle


print('Setting up GUI...')

with app.frame('loading', row=1, colspan=3):
    app.addLabel('l-label', 'Getting ready...', colspan=3)


def show_unknowntype(path: str):
    app.warningBox('ninfs Error',
                   "The type of the given file couldn't be detected."
                   "If you know it is a compatibile file, choose the "
                   "correct type and file an issue on GitHub if it works.\n\n"
                   + path)


def detect_type(fn: str):
    if fn.startswith('{'):
        fn = fn[1:-1]
    # noinspection PyBroadException
    try:
        with open(fn, 'rb') as f:
            mt = detect_format(f.read(0x400))
            if mt is not None:
                mount_type = mount_types_rv[mt]
            else:
                show_unknowntype(fn)
                return
    except (IsADirectoryError, PermissionError):  # PermissionError sometimes occurs on Windows
        if isfile(pjoin(fn, 'tmd')):
            mount_type = CDN
        else:
            try:
                next(iglob(pjoin(fn, '[0-9a-f]' * 32)))
            except StopIteration:
                # no entries
                mount_type = TITLEDIR
            else:
                # at least one entry
                mount_type = SD
    except Exception:
        print('Failed to get type of', fn)
        print_exc()
        return

    app.setOptionBox('TYPE', mount_type)
    app.setEntry(mount_type + ITEM, fn)
    app.showFrame(mount_type)


app.setSticky('new')
# this is here to force the width to be set
app.addOptionBox('TYPE', ('- Choose a type or drag a file/directory here -',), row=0, colspan=2)
app.setOptionBoxChangeFunction('TYPE', change_type)
app.addButton('Settings & Help', press, row=0, column=2)
try:
    app.setOptionBoxDropTarget('TYPE', detect_type)
    has_dnd = True
except Exception as e:
    print('Warning: Failed to enable Drag & Drop, will not be used.')
    print_exc()
    has_dnd = False

app.changeOptionBox('TYPE', (f'- Choose a type{" or drag a file/directory here" if has_dnd else ""} -',
                             '- Nintendo 3DS -', *ctr_types,
                             '- Nintendo DS / DSi -', *twl_types,
                             '- Nintendo Switch -', *hac_types),)

with app.frame('default', row=1, colspan=3):
    app.setSticky(EASTWEST)
    with app.labelFrame('Getting started', colspan=3):
        app.setSticky(EASTWEST)
        if has_dnd:
            app.addLabel('d-label1', 'To get started, drag the file to the box above.', colspan=2)
            app.addLabel('d-label2', 'You can also click it to manually choose a type.', colspan=2)
        else:
            app.addLabel('d-label1', 'To get started, choose a type to mount above.', colspan=2)
        app.addHorizontalSeparator(colspan=2)
        app.addLabel('d-label3', 'Need help?')
        app.addWebLink('View a tutorial here!', 'https://gbatemp.net/threads/499994/', column=1, row=PV)
        app.setLabelAlign('d-label3', 'right')
        app.setLinkAlign('View a tutorial here!', 'left')
app.hideFrame('default')  # to be shown later


def select_seeddb(sw):
    global seeddb_found
    path: str = app.openBox(title='Choose SeedDB', fileTypes=())
    if path:
        if getsize(path) % 0x10:
            app.warningBox('ninfs Error',
                           'The size for the selected SeedDB is not aligned to 0x10.')
            # shouldn't be calling this directly but i seem to have encounered a bug with parent=
            # noinspection PyProtectedMember
            app._bringToFront(sw)
            return
        with open(path, 'rb') as f:
            data = f.read(0x1FFFF0)
            target = pjoin(target_dir, 'seeddb.bin')
            makedirs(target_dir, exist_ok=True)
            with open(target, 'wb') as o:
                o.write(data)
            app.infoBox('ninfs', 'SeedDB was copied to:\n\n' + target)
            seeddb_found = target
            with suppress(ItemLookupError):
                app.hideLabel('no-seeddb')
                app.hideButton('fix-seeddb')
                if b9_found and seeddb_found:
                    app.hideFrame('FOOTER')
                app.hideSubWindow('no-seeddb')


if not b9_found or not seeddb_found:
    with app.frame('FOOTER', row=3, colspan=3):
        if not b9_found:
            app.addLabel('no-b9', 'boot9 was not found. Click for more details.', colspan=2)
            app.setLabelBg('no-b9', '#ff9999')
            app.addNamedButton('Fix boot9', 'fix-b9', lambda _: app.showSubWindow('no-b9'), row=PV, column=2)


            def select_b9(sw):
                global b9_found
                path: str = app.openBox(title='Choose boot9', fileTypes=(), parent='no-b9')
                if path:
                    if getsize(path) not in {0x8000, 0x10000}:
                        app.warningBox('ninfs Error',
                                       'The size for the selected boot9 match is not 0x8000 or 0x10000.')
                        # shouldn't be calling this directly but i seem to have encounered a bug with parent=
                        # noinspection PyProtectedMember
                        app._bringToFront(sw)
                        return
                    with open(path, 'rb') as f:
                        data = f.read(0x10000)
                        try:
                            fn = b9_hashes[sha256(data).hexdigest()]
                        except KeyError:
                            app.warningBox('ninfs Error', 'boot9 hash did not match. Please re-dump boot9.')
                            # noinspection PyProtectedMember
                            app._bringToFront(sw)
                            return
                        target = pjoin(target_dir, fn + '.bin')
                        makedirs(target_dir, exist_ok=True)
                        with open(target, 'wb') as o:
                            o.write(data)
                        app.infoBox('ninfs', 'boot9 was copied to:\n\n' + target)
                        b9_found = target
                        app.hideLabel('no-b9')
                        app.hideButton('fix-b9')
                        if b9_found and seeddb_found:
                            app.hideFrame('FOOTER')
                        check_mount_button(app.getOptionBox('TYPE'))
                        app.hideSubWindow('no-b9')


            with app.subWindow('no-b9', 'ninfs Error', modal=True) as sw:
                app.addLabel('boot9 was not found. It is needed for encryption.\n'
                             'Mount types that use encryption have been disabled.\n'
                             '\n'
                             'Choose this to automatically set up boot9.')
                app.addNamedButton('Select boot9 to copy...', 'no-b9-select', lambda _: select_b9(sw))
                app.setSticky(EASTWEST)
                app.addHorizontalSeparator()
                app.addLabel('The following paths were checked for a boot9 dump:')
                app.addListBox('b9-paths', b9_paths)
                app.setListBoxRows('b9-paths', len(b9_paths))
                app.setSticky('')
                app.addNamedButton(OK, 'no-b9-ok', lambda _: app.hideSubWindow('no-b9'))
                app.setResizable(False)

        if not seeddb_found:
            app.addLabel('no-seeddb', 'SeedDB was not found. Click for more details.', colspan=2)
            app.setLabelBg('no-seeddb', '#ffff99')
            app.addNamedButton('Fix SeedDB', 'fix-seeddb', lambda _: app.showSubWindow('no-seeddb'), row=PV,
                               column=2)

            with app.subWindow('no-seeddb', 'ninfs Error', modal=True) as sw:
                app.addLabel('SeedDB was not found. It is needed for encryption\n'
                             'of newer digital titles.\n'
                             '\n'
                             'Choose this to automatically set up SeedDB.')
                app.addNamedButton('Select SeedDB to copy...', 'no-seeddb-select', lambda _: select_seeddb(sw))
                app.setSticky(EASTWEST)
                app.addHorizontalSeparator()
                app.addLabel('The following paths were checked for SeedDB:')
                app.addListBox('seeddb-paths', seeddb_paths)
                app.setListBoxRows('seeddb-paths', len(seeddb_paths))
                app.setSticky('')
                app.addNamedButton(OK, 'no-seeddb-ok', lambda _: app.hideSubWindow('no-seeddb'))
                app.setResizable(False)
    app.hideFrame('FOOTER')

if windows:
    app.setFont(10)
elif macos:
    app.setFont(14)
app.setResizable(False)


# failed to mount subwindow
def show_mounterror():
    app.warningBox('ninfs Error', 'Failed to mount. Please check the output.')


# exited with error subwindow
def show_exiterror(errcode: int):
    app.warningBox('ninfs Error',
                   f'The mount process exited with an error code ({errcode}). Please check the output.')


# failed to mount to directory subwindow
if windows:
    def show_mounterror_dir_win():
        app.warningBox('ninfs Error',
                       'Failed to mount to the given mount point.\n'
                       'Please make sure the directory is empty or does not exist.')


def show_extras():
    try:
        app.showSubWindow('extras')
    except ItemLookupError:
        with app.subWindow('extras', 'ninfs Extras', modal=True, blocking=False) as sw:
            app.setSticky(EASTWEST)
            with app.labelFrame('Settings'):
                def checkbox_update(name: str):
                    if name == 'update-check':
                        option = app.getCheckBox('update-check')
                        if not write_config('update', 'update', 'check_updates_online', option):
                            app.errorBox('ninfs Error', 'Failed to write to config. Is the path read-only? '
                                                           'Check the output for more details.')

                app.addNamedCheckBox('Check for updates at launch', 'update-check')
                app.setCheckBox('update-check', update_config.getboolean('update', 'check_updates_online'))
                app.setCheckBoxChangeFunction('update-check', checkbox_update)

            app.setSticky(EASTWEST)
            with app.labelFrame('Update SeedDB', colspan=3):
                app.setSticky(EASTWEST)
                app.addLabel('updateseeddb-label', 'Update SeedDB to a newer database.', colspan=2)
                app.addNamedButton('Update', 'updateseeddb-btn', lambda _: select_seeddb(sw), row=PV, column=2)

            app.setSticky(EASTWEST)
            with app.labelFrame('Tutorial', colspan=3):
                app.setSticky(EASTWEST)
                app.addLabel('tutorial-label', 'View a tutorial on GBAtemp.', colspan=2)
                app.addNamedButton('Open', 'tutorial-btn',
                                   lambda _: webbrowser.open('https://gbatemp.net/threads/499994/'),
                                   row=PV, column=2)

            def _show_ctxmenu_window():
                app.showSubWindow('ctxmenu-window')
                app.hideSubWindow('extras')

            if windows:
                app.setSticky(EASTWEST)
                with app.labelFrame('Context Menu', colspan=3):
                    app.setSticky(EASTWEST)
                    app.addLabel('ctxmenu-label', 'Add an entry to the right-click menu.', colspan=2)
                    app.addNamedButton('Add', 'ctxmenu-btn', _show_ctxmenu_window, row=PV,
                                       column=2)

            def _show_about():
                app.showSubWindow('about')
                app.hideSubWindow('extras')

            app.setSticky(EASTWEST)
            with app.labelFrame('About', colspan=3):
                app.setSticky(EASTWEST)
                app.addLabel('about-label', 'Open the about dialog.')
                app.addNamedButton('Open', 'about-btn', _show_about, row=PV, column=2)

            app.setResizable(False)

        app.setSticky(EASTWEST)
        with app.subWindow('about', 'ninfs', modal=True, blocking=False):
            app.setSticky(EASTWEST)
            if windows:
                _ver = win32_ver()
                os_ver = 'Windows ' + _ver[0]
                if _ver[0] == '10':
                    k = OpenKey(HKEY_LOCAL_MACHINE, r'SOFTWARE\Microsoft\Windows NT\CurrentVersion')
                    try:
                        r: str = QueryValueEx(k, 'ReleaseId')[0]
                    except FileNotFoundError:
                        r = '1507'  # assuming RTM, since this key only exists in 1511 and up
                    os_ver += ', version ' + r
                else:
                    if _ver[2] != 'SP0':
                        os_ver += ' ' + _ver[2]
            elif macos:
                os_ver = 'macOS ' + mac_ver()[0]
            else:
                os_ver = ''
            if os_ver:
                os_ver += '\n'
            app.addMessage('about-msg', f'ninfs v{version}\n'
                                        f'Running on Python {pyver} {pybits}-bit\n'
                                        f'{os_ver}'
                                        f'\n'
                                        f'ninfs is released under the MIT license.', colspan=4)
            app.setMessageAspect('about-msg', 500)
            app.addWebLink('View ninfs on GitHub', 'https://github.com/ihaveamac/ninfs', colspan=4)
            app.addLabel('These libraries are used in the project:', colspan=4)
            app.addWebLink('appJar', 'https://github.com/jarvisteach/appJar')
            app.addWebLink('PyCryptodome', 'https://github.com/Legrandin/pycryptodome', row=PV, column=1)
            app.addWebLink('fusepy', 'https://github.com/fusepy/fusepy', row=PV, column=2)
            app.addWebLink('PyInstaller', 'https://github.com/pyinstaller/pyinstaller', row=PV, column=3)
            app.setResizable(False)

        if windows:
            def add_entry(button: str):
                app.hideSubWindow('ctxmenu-window')
                if button == 'Add entry':
                    add_reg(_used_pyinstaller)
                elif button == 'Remove entry':
                    del_reg()

            with app.subWindow('ctxmenu-window', 'ninfs', modal=True):
                msg = (
                    'A new entry can be added to the Windows context menu when you right-click on a file, providing an '
                    'easy way to mount various files in Windows Explorer using ninfs.\n'
                    '\n'
                    'This will modify the registry to add it.')
                if _used_pyinstaller:
                    msg += ' If you move or rename the EXE, you will need to re-add the entry.'
                app.addMessage('extras-ctxmenu-label', msg)
                app.setMessageAspect('extras-ctxmenu-label', 300)
                app.addButtons(['Add entry', 'Remove entry', 'Cancel'], add_entry, colspan=3)
                app.setResizable(False)

        app.showSubWindow('extras')


# file/directory not set error
def show_noitemerror():
    app.infoBox('ninfs Error', 'Select a file or directory to mount.')


# mountpoint not set error
def show_nomperror():
    app.infoBox('ninfs Error', 'Select an empty directory to be the mount point.')


# failed to unmount subwindow
with app.subWindow('unmounterror', 'ninfs Error', modal=True, blocking=False):
    def unmount_ok(_):
        app.hideSubWindow('unmounterror')
        app.enableButton(UNMOUNT)


    app.addLabel('unmounterror-label', 'Failed to unmount. Please check the output.\n\n'
                                       'You can kill the process if it is not responding. '
                                       'This should be used as a last resort. '
                                       'The process should be unmounted normally.', colspan=2)
    app.addNamedButton(OK, 'unmounterror-ok', unmount_ok)
    app.addNamedButton('Kill process', 'unmounterror-kill', kill_process, row=PV, column=1)
    app.setResizable(False)


# maybe get the file via an argument to main? instead of taking it from argv
# it probably doesn't matter much
def main(_pyi=False, _allow_admin=False):
    global _used_pyinstaller
    _used_pyinstaller = _pyi
    print('Doing final setup...')
    try:
        # attempt importing all the fusepy stuff used in the mount scripts
        # if it fails, libfuse probably couldn't be found
        import fuse
    except EnvironmentError:
        # TODO: probably check if this was really "Unable to find libfuse" (this is aliased to OSError)
        if windows:
            res = app.yesNoBox('ninfs',
                               'Failed to import fusepy. WinFsp needs to be installed.\n\n'
                               'Would you like to open the WinFsp download page?\n'
                               'http://www.secfs.net/winfsp/download/')
            if res:
                webbrowser.open('http://www.secfs.net/winfsp/download/')
        elif macos:
            res = app.yesNoBox('ninfs',
                               'Failed to import fusepy. FUSE for macOS needs to be installed.\n\n'
                               'Would you like to open the FUSE for macOS download page?\n'
                               'https://osxfuse.github.io')
            if res:
                webbrowser.open('https://osxfuse.github.io')
        else:
            print("Failed to load fusepy. libfuse probably couldn't be found.")
        return 1

    if windows and not _allow_admin:
        isadmin: int = windll.shell32.IsUserAnAdmin()
        if isadmin and uac_enabled():
            app.warningBox('ninfs',
                           'This should not be run as administrator.\n'
                           'The mount point may not be accessible by your account normally, '
                           'only by the administrator.\n\n'
                           'If you are having issues with administrative tools not seeing files, '
                           'choose a directory as a mount point instead of a drive letter.')
            exit(1)

    def show_update(name: str, ver: str, info: str, url: str):
        def update_press(button: str):
            if button == 'Open release page':
                webbrowser.open(url)
                app.queueFunction(app.stop)
            elif button == 'Ignore this update':
                if not write_config('update', 'update', 'ignored_update', ver):
                    app.errorBox('ninfs Error', 'Failed to write to config. Is the path read-only? '
                                                   'Check the output for more details.')
            app.destroySubWindow('update')

        with app.subWindow('update', 'ninfs Update', modal=True):
            app.addLabel('update-label1', f'A new version of ninfs is available. You have v{version}.')
            app.addButtons(['Open release page', 'Ignore this update', 'Close'], update_press)
            with app.labelFrame(name):
                app.addMessage('update-info', info)
                app.setMessageAlign('update-info', 'left')
                app.setMessageAspect('update-info', 300)
            app.setResizable(False)

        app.showSubWindow('update')

    if update_config.getboolean('update', 'check_updates_online'):
        def update_check():
            # this will check for the latest non-prerelease, once there is one.
            # noinspection PyBroadException
            try:
                print(f'UPDATE: Checking for updates... (Currently running v{version})')
                ctx = SSLContext(PROTOCOL_TLSv1_2)
                release_url = 'https://api.github.com/repos/ihaveamac/ninfs/releases'
                current_ver: 'Version' = parse_version(version)
                if not current_ver.is_prerelease:
                    release_url += '/latest'
                with urlopen(release_url, context=ctx) as u:
                    u: HTTPResponse
                    res: List[Dict[str, Any]] = json.loads(u.read().decode('utf-8'))
                    latest_rel = res[0] if current_ver.is_prerelease else res
                    latest_ver: str = latest_rel['tag_name']
                    if parse_version(latest_ver) > current_ver:
                        name: str = latest_rel['name']
                        url: str = latest_rel['html_url']
                        info_all: str = latest_rel['body']
                        info = info_all[:info_all.find('------')].strip().replace('\r\n', '\n')

                        if latest_ver == update_config['update']['ignored_update']:
                            print(f'UPDATE: Update to {latest_ver} is available but ignored.')
                        else:
                            print(f'UPDATE: Update to {latest_ver} is available.')
                            app.queueFunction(show_update, name, latest_ver, info, url)
                    else:
                        print(f'UPDATE: No new version. (Latest is {latest_ver})')

            except Exception:
                print('UPDATE: Failed to check for update')
                print_exc()

        app.thread(update_check)
    else:
        print('UPDATE: Online update check disabled.')

    to_use = 'default'

    if len(argv) > 1:
        fn: str = abspath(argv[1])
        # noinspection PyBroadException
        try:
            with open(fn, 'rb') as f:
                mt = detect_format(f.read(0x400))
                if mt is not None:
                    mount_type = mount_types_rv[mt]
                    to_use = mount_type
                    file_to_use = fn
                else:
                    show_unknowntype(fn)
        except (IsADirectoryError, PermissionError):  # PermissionError sometimes occurs on Windows
            if isfile(pjoin(fn, 'tmd')):
                mount_type = CDN
            else:
                try:
                    next(iglob(pjoin(fn, '[0-9a-f]' * 32)))
                except StopIteration:
                    # no entries
                    mount_type = TITLEDIR
                else:
                    # at least one entry
                    mount_type = SD
            to_use = mount_type
            file_to_use = fn
        except Exception:
            print('Failed to get type of', fn)
            print_exc()

    # kinda lame way to prevent a resize bug
    def sh():
        if to_use == 'default':
            app.queueFunction(app.showFrame, 'default')
        else:
            app.queueFunction(detect_type, file_to_use)
        app.queueFunction(app.hideFrame, 'loading')

    app.thread(sh)

    print('Starting the GUI!')
    app.go()
    stop_mount()
    return 0



# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

from functools import wraps
from io import BufferedIOBase, TextIOWrapper
from threading import Lock
from typing import overload, TYPE_CHECKING, NamedTuple

from ..common import PyCTRError
from ..util import readle, roundup

if TYPE_CHECKING:
    from typing import BinaryIO, Optional, Tuple, Union

__all__ = ['IVFC_HEADER_SIZE', 'IVFC_ROMFS_MAGIC_NUM', 'ROMFS_LV3_HEADER_SIZE', 'RomFSError', 'InvalidIVFCError',
           'InvalidRomFSHeaderError', 'RomFSEntryError', 'RomFSFileNotFoundError', 'RomFSReader']

IVFC_HEADER_SIZE = 0x5C
IVFC_ROMFS_MAGIC_NUM = 0x10000
ROMFS_LV3_HEADER_SIZE = 0x28


class RomFSError(PyCTRError):
    """Generic exception for RomFS operations."""


class InvalidIVFCError(RomFSError):
    """Invalid IVFC header exception."""


class InvalidRomFSHeaderError(RomFSError):
    """Invalid RomFS Level 3 header."""


class RomFSEntryError(RomFSError):
    """Error with RomFS Directory or File entry."""


class RomFSFileNotFoundError(RomFSError):
    """Invalid file path in RomFS Level 3."""


class RomFSRegion(NamedTuple):
    offset: int
    size: int


class RomFSDirectoryEntry(NamedTuple):
    name: str
    type: str
    contents: 'Tuple[str, ...]'


class RomFSFileEntry(NamedTuple):
    name: str
    type: str
    offset: int
    size: int


def _raise_if_closed(method):
    @wraps(method)
    def decorator(self: '_RomFSOpenFile', *args, **kwargs):
        if self._reader.closed:
            self.closed = True
        if self.closed:
            raise ValueError('I/O operation on closed file.')
        return method(self, *args, **kwargs)
    return decorator


class _RomFSOpenFile(BufferedIOBase):
    """Class for open RomFS file entries."""

    _seek = 0

    # noinspection PyMissingConstructor
    def __init__(self, reader: 'RomFSReader', path: str):
        self._reader = reader
        self._path = path
        self._info: RomFSFileEntry = reader.get_info_from_path(path)
        if not isinstance(self._info, RomFSFileEntry):
            raise RomFSEntryError('not a file entry: ' + path)

    def __repr__(self) -> str:
        return f'<_RomFSOpenFile path={self._path!r} info={self._info!r} reader={self._reader!r}>'

    @_raise_if_closed
    def read(self, size: int = -1) -> bytes:
        if size == -1:
            size = self._info.size - self._seek
        data = self._reader.get_data(self._info, self._seek, size)
        self._seek += len(data)
        return data

    read1 = read  # probably make this act like read1 should, but this for now enables some other things to work

    @_raise_if_closed
    def seek(self, seek: int, whence: int = 0) -> int:
        if whence == 0:
            if seek < 0:
                raise ValueError(f'negative seek value {seek}')
            self._seek = min(seek, self._info.size)
        elif whence == 1:
            self._seek = max(self._seek + seek, 0)
        elif whence == 2:
            self._seek = max(self._info.size + seek, 0)
        return self._seek

    @_raise_if_closed
    def tell(self) -> int:
        return self._seek

    @_raise_if_closed
    def readable(self) -> bool:
        return True

    @_raise_if_closed
    def seekable(self) -> bool:
        return True


class RomFSReader:
    """
    Class for 3DS RomFS Level 3 partition.

    https://www.3dbrew.org/wiki/RomFS
    """

    closed = False
    lv3_offset = 0
    data_offset = 0

    def __init__(self, fp: 'Union[str, BinaryIO]', case_insensitive: bool = False):
        if isinstance(fp, str):
            fp = open(fp, 'rb')
        self._fp = fp
        self.case_insensitive = case_insensitive
        self._lock = Lock()

        lv3_offset = fp.tell()
        magic = fp.read(4)

        # detect ivfc and get the lv3 offset
        if magic == b'IVFC':
            ivfc = magic + fp.read(0x54)  # IVFC_HEADER_SIZE - 4
            ivfc_magic_num = readle(ivfc[0x4:0x8])
            if ivfc_magic_num != IVFC_ROMFS_MAGIC_NUM:
                raise InvalidIVFCError(f'IVFC magic number is invalid '
                                       f'({ivfc_magic_num:#X} instead of {IVFC_ROMFS_MAGIC_NUM:#X})')
            master_hash_size = readle(ivfc[0x8:0xC])
            lv3_block_size = readle(ivfc[0x4C:0x50])
            lv3_hash_block_size = 1 << lv3_block_size
            lv3_offset += roundup(0x60 + master_hash_size, lv3_hash_block_size)
            fp.seek(lv3_offset)
            magic = fp.read(4)
        self.lv3_offset = lv3_offset

        lv3_header = magic + fp.read(0x24)  # ROMFS_LV3_HEADER_SIZE - 4

        # get offsets and sizes from lv3 header
        lv3_header_size = readle(magic)
        lv3_dirhash = RomFSRegion(offset=readle(lv3_header[0x4:0x8]), size=readle(lv3_header[0x8:0xC]))
        lv3_dirmeta = RomFSRegion(offset=readle(lv3_header[0xC:0x10]), size=readle(lv3_header[0x10:0x14]))
        lv3_filehash = RomFSRegion(offset=readle(lv3_header[0x14:0x18]), size=readle(lv3_header[0x18:0x1C]))
        lv3_filemeta = RomFSRegion(offset=readle(lv3_header[0x1C:0x20]), size=readle(lv3_header[0x20:0x24]))
        lv3_filedata_offset = readle(lv3_header[0x24:0x28])
        self.data_offset = lv3_offset + lv3_filedata_offset

        # verify lv3 header
        if lv3_header_size != ROMFS_LV3_HEADER_SIZE:
            raise InvalidRomFSHeaderError('Length in RomFS Lv3 header is not 0x28')
        if lv3_dirhash.offset < lv3_header_size:
            raise InvalidRomFSHeaderError('Directory Hash offset is before the end of the Lv3 header')
        if lv3_dirmeta.offset < lv3_dirhash.offset + lv3_dirhash.size:
            raise InvalidRomFSHeaderError('Directory Metadata offset is before the end of the Directory Hash region')
        if lv3_filehash.offset < lv3_dirmeta.offset + lv3_dirmeta.size:
            raise InvalidRomFSHeaderError('File Hash offset is before the end of the Directory Metadata region')
        if lv3_filemeta.offset < lv3_filehash.offset + lv3_filehash.size:
            raise InvalidRomFSHeaderError('File Metadata offset is before the end of the File Hash region')
        if lv3_filedata_offset < lv3_filemeta.offset + lv3_filemeta.size:
            raise InvalidRomFSHeaderError('File Data offset is before the end of the File Metadata region')

        # get entries from dirmeta and filemeta
        def iterate_dir(out: dict, raw: bytes, current_path: str):
            first_child_dir = readle(raw[0x8:0xC])
            first_file = readle(raw[0xC:0x10])

            out['type'] = 'dir'
            out['contents'] = {}

            # iterate through all child directories
            if first_child_dir != 0xFFFFFFFF:
                fp.seek(lv3_offset + lv3_dirmeta.offset + first_child_dir)
                while True:
                    child_dir_meta = fp.read(0x18)
                    next_sibling_dir = readle(child_dir_meta[0x4:0x8])
                    child_dir_name = fp.read(readle(child_dir_meta[0x14:0x18])).decode('utf-16le')
                    child_dir_name_meta = child_dir_name.lower() if case_insensitive else child_dir_name
                    if child_dir_name_meta in out['contents']:
                        print(f'WARNING: Dirname collision! {current_path}{child_dir_name}')
                    out['contents'][child_dir_name_meta] = {'name': child_dir_name}

                    iterate_dir(out['contents'][child_dir_name_meta], child_dir_meta,
                                f'{current_path}{child_dir_name}/')
                    if next_sibling_dir == 0xFFFFFFFF:
                        break
                    fp.seek(lv3_offset + lv3_dirmeta.offset + next_sibling_dir)

            if first_file != 0xFFFFFFFF:
                fp.seek(lv3_offset + lv3_filemeta.offset + first_file)
                while True:
                    child_file_meta = fp.read(0x20)
                    next_sibling_file = readle(child_file_meta[0x4:0x8])
                    child_file_offset = readle(child_file_meta[0x8:0x10])
                    child_file_size = readle(child_file_meta[0x10:0x18])
                    child_file_name = fp.read(readle(child_file_meta[0x1C:0x20])).decode('utf-16le')
                    child_file_name_meta = child_file_name.lower() if self.case_insensitive else child_file_name
                    if child_file_name_meta in out['contents']:
                        print(f'WARNING: Filename collision! {current_path}{child_file_name}')
                    out['contents'][child_file_name_meta] = {'name': child_file_name, 'type': 'file',
                                                             'offset': child_file_offset, 'size': child_file_size}

                    self.total_size += child_file_size
                    if next_sibling_file == 0xFFFFFFFF:
                        break
                    fp.seek(lv3_offset + lv3_filemeta.offset + next_sibling_file)

        self._tree_root = {'name': 'ROOT'}
        self.total_size = 0
        fp.seek(lv3_offset + lv3_dirmeta.offset)
        iterate_dir(self._tree_root, fp.read(0x18), '/')

    def close(self):
        self.closed = True
        self._fp.close()

    @overload
    def open(self, path: str, encoding: str, errors: 'Optional[str]' = None,
             newline: 'Optional[str]' = None) -> TextIOWrapper: ...

    @overload
    def open(self, path: str, encoding: None = None, errors: 'Optional[str]' = None,
             newline: 'Optional[str]' = None) -> _RomFSOpenFile: ...

    def open(self, path, encoding=None, errors=None, newline=None):
        """Open a file in the RomFS for reading."""
        f = _RomFSOpenFile(self, path)
        if encoding is not None:
            f = TextIOWrapper(f, encoding, errors, newline)
        return f

    __del__ = close

    def get_info_from_path(self, path: str) -> 'Union[RomFSDirectoryEntry, RomFSFileEntry]':
        """Get a directory or file entry"""
        curr = self._tree_root
        if self.case_insensitive:
            path = path.lower()
        if path[0] == '/':
            path = path[1:]
        for part in path.split('/'):
            if part == '':
                break
            try:
                # noinspection PyTypeChecker
                curr = curr['contents'][part]
            except KeyError:
                raise RomFSFileNotFoundError(path)
        if curr['type'] == 'dir':
            contents = (k['name'] for k in curr['contents'].values())
            return RomFSDirectoryEntry(name=curr['name'], type='dir', contents=(*contents,))
        elif curr['type'] == 'file':
            return RomFSFileEntry(name=curr['name'], type='file', offset=curr['offset'], size=curr['size'])

    def get_data(self, info: RomFSFileEntry, offset: int, size: int) -> bytes:
        if offset + size > info.size:
            size = info.size - offset
        with self._lock:
            self._fp.seek(self.data_offset + info.offset + offset)
            return self._fp.read(size)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

"""
Mounts SD contents under `/Nintendo 3DS`, creating a virtual filesystem with decrypted contents. movable.sed required.
"""

import logging
import os
from contextlib import suppress
from errno import EPERM, EACCES, EBADF
from hashlib import sha256
from struct import unpack
from sys import exit, argv
from typing import TYPE_CHECKING

from pyctr.crypto import CryptoEngine, Keyslot
from pyctr.util import readbe
from . import _common as _c
# _common imports these from fusepy, and prints an error if it fails; this allows less duplicated code
from ._common import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context

if TYPE_CHECKING:
    from typing import Dict, BinaryIO

if _c.windows:
    from ctypes import c_wchar_p, pointer, c_ulonglong, windll, wintypes


class SDFilesystemMount(LoggingMixIn, Operations):
    @_c.ensure_lower_path
    def path_to_iv(self, path):
        path_hash = sha256(path[self.root_len + 33:].encode('utf-16le') + b'\0\0').digest()
        hash_p1 = readbe(path_hash[0:16])
        hash_p2 = readbe(path_hash[16:32])
        return hash_p1 ^ hash_p2

    def __init__(self, sd_dir: str, movable: str, dev: bool = False, readonly: bool = False):
        self.crypto = CryptoEngine(dev=dev)

        with open(movable, 'rb') as mv:
            mv.seek(0x110)
            key_y = mv.read(0x10)
        key_hash = sha256(key_y).digest()
        hash_parts = unpack('<IIII', key_hash[0:16])
        self.root_dir = f'{hash_parts[0]:08x}{hash_parts[1]:08x}{hash_parts[2]:08x}{hash_parts[3]:08x}'

        if not os.path.isdir(sd_dir + '/' + self.root_dir):
            exit(f'Failed to find {self.root_dir} in the SD dir.')

        self.fds: Dict[int, BinaryIO] = {}

        print('Root dir: ' + self.root_dir)

        self.crypto.set_keyslot('y', Keyslot.SD, readbe(key_y))
        print('Key:      ' + self.crypto.key_normal[Keyslot.SD].hex())

        self.root = os.path.realpath(sd_dir + '/' + self.root_dir)
        self.root_len = len(self.root)

        self.readonly = readonly

    # noinspection PyMethodOverriding
    def __call__(self, op, path, *args):
        return super().__call__(op, self.root + path, *args)

    def __del__(self, *args):
        # putting the keys in a tuple so the dict can be modified
        with suppress(AttributeError):
            for f in tuple(self.fds):
                with suppress(KeyError):
                    self.fds[f].close()
                    del self.fds[f]

    destroy = __del__

    def access(self, path, mode):
        if not os.access(path, mode):
            raise FuseOSError(EACCES)

    @_c.raise_on_readonly
    def chmod(self, path, mode):
        os.chmod(path, mode)

    @_c.raise_on_readonly
    def chown(self, path, *args, **kwargs):
        if not _c.windows:
            os.chown(path, *args, **kwargs)

    @_c.raise_on_readonly
    def create(self, path, mode, **kwargs):
        fd = os.open(path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, mode)
        self.fds[fd] = os.fdopen(fd, 'wb')
        return fd

    def flush(self, path, fh):
        try:
            os.fsync(fh)
        except OSError as e:
            # I am not sure why this is happening on Windows. if anyone can give me a hint, please do.
            if e.errno != EBADF:  # "Bad file descriptor"
                raise
        return

    def fsync(self, path, datasync, fh):
        self.flush(path, fh)
        return

    def getattr(self, path, fh=None):
        st = os.lstat(path)
        uid, gid, _ = fuse_get_context()
        res = {key: getattr(st, key) for key in ('st_atime', 'st_ctime', 'st_mode', 'st_mtime', 'st_nlink', 'st_size',
                                                 'st_flags') if hasattr(st, key)}
        res['st_uid'] = st.st_uid if st.st_uid != 0 else uid
        res['st_gid'] = st.st_gid if st.st_gid != 0 else gid
        return res

    getxattr = None

    def link(self, target, source):
        return os.link(source, target)

    listxattr = None

    @_c.raise_on_readonly
    def mkdir(self, path, *args, **kwargs):
        os.mkdir(path, *args, **kwargs)

    @_c.raise_on_readonly
    def mknod(self, path, *args, **kwargs):
        if not _c.windows:
            os.mknod(path, *args, **kwargs)

    # open = os.open
    def open(self, path, flags):
        f = open(path, 'rb+', buffering=-1)
        self.fds[f.fileno()] = f
        return f.fileno()

    def read(self, path, size, offset, fh):
        f = self.fds[fh]
        # special check for special files
        if os.path.basename(path).startswith('.') or 'nintendo dsiware' in path:
            f.seek(offset)
            return f.read(size)

        before = offset % 16
        f.seek(offset - before)
        data = f.read(size + before)
        iv = self.path_to_iv(path) + (offset >> 4)
        return self.crypto.create_ctr_cipher(Keyslot.SD, iv).decrypt(data)[before:]

    def readdir(self, path, fh):
        yield from ('.', '..')
        ld = os.listdir(path)
        if _c.windows:
            # I should figure out how to mark hidden files, if possible
            yield from (d for d in ld if not d.startswith('.'))
        else:
            yield from ld

    readlink = os.readlink

    def release(self, path, fh):
        self.fds[fh].close()
        del self.fds[fh]

    @_c.raise_on_readonly
    def rename(self, old, new):
        # renaming's too difficult. just copy the file to the name you want if you really need it.
        raise FuseOSError(EPERM)

    @_c.raise_on_readonly
    def rmdir(self, path):
        os.rmdir(path)

    # noinspection PyPep8Naming
    def statfs(self, path):
        if _c.windows:
            lpSectorsPerCluster = c_ulonglong(0)
            lpBytesPerSector = c_ulonglong(0)
            lpNumberOfFreeClusters = c_ulonglong(0)
            lpTotalNumberOfClusters = c_ulonglong(0)
            ret = windll.kernel32.GetDiskFreeSpaceW(c_wchar_p(path), pointer(lpSectorsPerCluster),
                                                    pointer(lpBytesPerSector),
                                                    pointer(lpNumberOfFreeClusters),
                                                    pointer(lpTotalNumberOfClusters))
            if not ret:
                raise WindowsError
            free_blocks = lpNumberOfFreeClusters.value * lpSectorsPerCluster.value
            result = {'f_bavail': free_blocks,
                      'f_bfree': free_blocks,
                      'f_bsize': lpBytesPerSector.value,
                      'f_frsize': lpBytesPerSector.value,
                      'f_blocks': lpTotalNumberOfClusters.value * lpSectorsPerCluster.value,
                      'f_namemax': wintypes.MAX_PATH}
            return result
        else:
            stv = os.statvfs(path)
            # f_flag causes python interpreter crashes in some cases. i don't get it.
            return {key: getattr(stv, key) for key in ('f_bavail', 'f_bfree', 'f_blocks', 'f_bsize', 'f_favail',
                                                       'f_ffree', 'f_files', 'f_frsize', 'f_namemax')}

    def symlink(self, target, source):
        return os.symlink(source, target)

    def truncate(self, path, length, fh=None):
        if fh is None:
            with open(path, 'r+b') as f:
                f.truncate(length)
        else:
            f = self.fds[fh]
            f.truncate(length)

    @_c.raise_on_readonly
    def unlink(self, path, *args, **kwargs):
        os.unlink(path)

    @_c.raise_on_readonly
    def utimens(self, path, *args, **kwargs):
        os.utime(path, *args, **kwargs)

    @_c.raise_on_readonly
    def write(self, path, data, offset, fh):
        f = self.fds[fh]
        # special check for special files
        if os.path.basename(path).startswith('.') or 'nintendo dsiware' in path.lower():
            f.seek(offset)
            return f.write(data)

        before = offset % 16
        iv = self.path_to_iv(path) + (offset >> 4)
        out_data = self.crypto.create_ctr_cipher(Keyslot.SD, iv).decrypt((b'\0' * before) + data)[before:]
        f.seek(offset)
        return f.write(out_data)


def main(prog: str = None, args: list = None):
    from argparse import ArgumentParser
    if args is None:
        args = argv[1:]
    parser = ArgumentParser(prog=prog, description='Mount Nintendo 3DS SD card contents.',
                            parents=(_c.default_argp, _c.readonly_argp, _c.dev_argp,
                                     _c.main_args(
                                         'sd_dir', "path to folder with SD contents (on SD: /Nintendo 3DS)")))
    parser.add_argument('--movable', metavar='MOVABLESED', help='path to movable.sed', required=True)

    a = parser.parse_args(args)
    opts = dict(_c.parse_fuse_opts(a.o))

    if a.do:
        logging.basicConfig(level=logging.DEBUG, filename=a.do)

    mount = SDFilesystemMount(sd_dir=a.sd_dir, movable=a.movable, dev=a.dev, readonly=a.ro)
    if _c.macos or _c.windows:
        opts['fstypename'] = 'SDCard'
        if _c.macos:
            opts['volname'] = f'Nintendo 3DS SD Card ({mount.root_dir})'
            opts['noappledouble'] = True  # fixes an error. but this is probably not the best way to do it.
        else:
            # windows
            opts['volname'] = f'Nintendo 3DS SD Card ({mount.root_dir[0:8]}…)'
            opts['case_insensitive'] = False
    FUSE(mount, a.mount_point, foreground=a.fg or a.do or a.d, ro=a.ro, nothreads=True, debug=a.d,
         fsname=os.path.realpath(a.sd_dir).replace(',', '_'), **opts)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

from hashlib import sha256
from struct import pack
from typing import TYPE_CHECKING, NamedTuple

from ..common import PyCTRError
from ..util import readbe

if TYPE_CHECKING:
    from typing import BinaryIO, Iterable

__all__ = ['CHUNK_RECORD_SIZE', 'TitleMetadataError', 'InvalidSignatureTypeError', 'InvalidHashError',
           'ContentInfoRecord', 'ContentChunkRecord', 'ContentTypeFlags', 'TitleVersion', 'TitleMetadataReader']

CHUNK_RECORD_SIZE = 0x30

# sig-type: (sig-size, padding)
signature_types = {
    # RSA_4096 SHA1 (unused on 3DS)
    0x00010000: (0x200, 0x3C),
    # RSA_2048 SHA1 (unused on 3DS)
    0x00010001: (0x100, 0x3C),
    # Elliptic Curve with SHA1 (unused on 3DS)
    0x00010002: (0x3C, 0x40),
    # RSA_4096 SHA256
    0x00010003: (0x200, 0x3C),
    # RSA_2048 SHA256
    0x00010004: (0x100, 0x3C),
    # ECDSA with SHA256
    0x00010005: (0x3C, 0x40),
}

BLANK_SIG_PAIR = (0x00010004, b'\xFF' * signature_types[0x00010004][0])


class TitleMetadataError(PyCTRError):
    """Generic exception for TitleMetadata operations."""


class InvalidTMDError(TitleMetadataError):
    """Title Metadata is invalid."""


class InvalidSignatureTypeError(InvalidTMDError):
    """Invalid signature type was used."""

    def __init__(self, sig_type):
        super().__init__(sig_type)

    def __str__(self):
        return f'{self.args[0]:#010x}'


class InvalidHashError(InvalidTMDError):
    """Hash mismatch in the Title Metadata."""


class InvalidInfoRecordError(InvalidHashError):
    """Hash mismatch in the Content Info Records."""

    def __init__(self, info_record):
        super().__init__(info_record)

    def __str__(self):
        return f'Invalid info record: {self.args[0]}'


class UnusualInfoRecordError(InvalidTMDError):
    """Encountered Content Info Record that attempts to hash a Content Chunk Record that has already been hashed."""

    def __init__(self, info_record, chunk_record):
        super().__init__(info_record, chunk_record)

    def __str__(self):
        return f'Attempted to hash twice: {self.args[0]}, {self.args[1]}'


class ContentTypeFlags(NamedTuple):
    encrypted: bool
    disc: bool
    cfm: bool
    optional: bool
    shared: bool

    def __index__(self) -> int:
        return self.encrypted | (self.disc << 1) | (self.cfm << 2) | (self.optional << 14) | (self.shared << 15)

    __int__ = __index__

    def __format__(self, format_spec: str) -> str:
        return self.__int__().__format__(format_spec)

    @classmethod
    def from_int(cls, flags: int) -> 'ContentTypeFlags':
        # noinspection PyArgumentList
        return cls(bool(flags & 1), bool(flags & 2), bool(flags & 4), bool(flags & 0x4000), bool(flags & 0x8000))


class ContentInfoRecord(NamedTuple):
    index_offset: int
    command_count: int
    hash: bytes

    def __bytes__(self) -> bytes:
        return b''.join((self.index_offset.to_bytes(2, 'big'), self.command_count.to_bytes(2, 'big'), self.hash))


class ContentChunkRecord(NamedTuple):
    id: str
    cindex: int
    type: ContentTypeFlags
    size: int
    hash: bytes

    def __bytes__(self) -> bytes:
        return b''.join((bytes.fromhex(self.id), self.cindex.to_bytes(2, 'big'), int(self.type).to_bytes(2, 'big'),
                         self.size.to_bytes(8, 'big'), self.hash))


class TitleVersion(NamedTuple):
    major: int
    minor: int
    micro: int

    def __str__(self) -> str:
        return f'{self.major}.{self.minor}.{self.micro}'

    def __index__(self) -> int:
        return (self.major << 10) | (self.minor << 4) | self.micro

    __int__ = __index__

    def __format__(self, format_spec: str) -> str:
        return self.__int__().__format__(format_spec)

    @classmethod
    def from_int(cls, ver: int) -> 'TitleVersion':
        # noinspection PyArgumentList
        return cls((ver >> 10) & 0x3F, (ver >> 4) & 0x3F, ver & 0xF)


class TitleMetadataReader:
    """
    Class for 3DS Title Metadata.

    https://www.3dbrew.org/wiki/Title_metadata
    """

    __slots__ = ('title_id', 'save_size', 'srl_save_size', 'title_version', 'info_records',
                 'chunk_records', 'content_count', 'signature', '_u_issuer', '_u_version', '_u_ca_crl_version',
                 '_u_signer_crl_version', '_u_reserved1', '_u_system_version', '_u_title_type', '_u_group_id',
                 '_u_reserved2', '_u_srl_flag', '_u_reserved3', '_u_access_rights', '_u_boot_count', '_u_padding')

    # arguments prefixed with _u_ are values unused by the 3DS and/or are only kept around to generate the final tmd
    def __init__(self, *, title_id: str, save_size: int, srl_save_size: int, title_version: TitleVersion,
                 info_records: 'Iterable[ContentInfoRecord]', chunk_records: 'Iterable[ContentChunkRecord]',
                 signature=BLANK_SIG_PAIR, _u_issuer='Root-CA00000003-CP0000000b', _u_version=1, _u_ca_crl_version=0,
                 _u_signer_crl_version=0, _u_reserved1=0, _u_system_version=b'\0' * 8, _u_title_type=b'\0\0\0@',
                 _u_group_id=b'\0\0', _u_reserved2=b'\0\0\0\0', _u_srl_flag=0, _u_reserved3=b'\0' * 0x31,
                 _u_access_rights=b'\0' * 4, _u_boot_count=b'\0\0', _u_padding=b'\0\0'):
        # TODO: add checks
        self.title_id = title_id.lower()
        self.save_size = save_size
        self.srl_save_size = srl_save_size
        self.title_version = title_version
        self.info_records = tuple(info_records)
        self.chunk_records = tuple(chunk_records)
        self.content_count = len(self.chunk_records)
        self.signature = signature  # TODO: store this differently

        # unused values
        self._u_issuer = _u_issuer
        self._u_version = _u_version
        self._u_ca_crl_version = _u_ca_crl_version
        self._u_signer_crl_version = _u_signer_crl_version
        self._u_reserved1 = _u_reserved1
        self._u_system_version = _u_system_version
        self._u_title_type = _u_title_type
        self._u_group_id = _u_group_id
        self._u_reserved2 = _u_reserved2
        self._u_srl_flag = _u_srl_flag
        self._u_reserved3 = _u_reserved3
        self._u_access_rights = _u_access_rights
        self._u_boot_count = _u_boot_count
        self._u_padding = _u_padding

    def __hash__(self) -> int:
        return hash((self.title_id, self.save_size, self.srl_save_size, self.title_version,
                     self.info_records, self.chunk_records))

    def __repr__(self) -> str:
        return (f'<TitleMetadataReader title_id={self.title_id!r} title_version={self.title_version!r} '
                f'content_count={self.content_count!r}>')

    def __bytes__(self) -> bytes:
        sig_data = pack(f'>I {signature_types[self.signature[0]][0]}s {signature_types[self.signature[0]][1]}x',
                        self.signature[0], self.signature[1])

        info_records = b''.join(bytes(x) for x in self.info_records).ljust(0x900, b'\0')

        header = pack('>64s b b b b 8s 8s 4s 2s I I 4s b 49s 4s H H 2s 2s 32s', self._u_issuer.encode('ascii'),
                      self._u_version, self._u_ca_crl_version, self._u_signer_crl_version, self._u_reserved1,
                      self._u_system_version, bytes.fromhex(self.title_id), self._u_title_type, self._u_group_id,
                      self.save_size, self.srl_save_size, self._u_reserved2, self._u_srl_flag, self._u_reserved3,
                      self._u_access_rights, self.title_version, self.content_count, self._u_boot_count,
                      self._u_padding, sha256(info_records).digest())

        chunk_records = b''.join(bytes(x) for x in self.chunk_records)

        return sig_data + header + info_records + chunk_records

    @classmethod
    def load(cls, fp: 'BinaryIO', verify_hashes: bool = True) -> 'TitleMetadataReader':
        """Load a tmd from a file-like object."""
        sig_type = readbe(fp.read(4))
        try:
            sig_size, sig_padding = signature_types[sig_type]
        except KeyError:
            raise InvalidSignatureTypeError(sig_type)

        signature = fp.read(sig_size)
        try:
            fp.seek(sig_padding, 1)
        except Exception:
            # most streams are probably seekable, but for some that aren't...
            fp.read(sig_padding)

        header = fp.read(0xC4)
        if len(header) != 0xC4:
            raise InvalidTMDError('Header length is not 0xC4')

        # only values that actually have a use are loaded here. (currently)
        # several fields in were left in from the Wii tmd and have no function on 3DS.
        title_id = header[0x4C:0x54].hex()
        save_size = readbe(header[0x5A:0x5E])
        srl_save_size = readbe(header[0x5E:0x62])
        title_version = TitleVersion.from_int(readbe(header[0x9C:0x9E]))
        content_count = readbe(header[0x9E:0xA0])

        content_info_records_hash = header[0xA4:0xC4]

        content_info_records_raw = fp.read(0x900)
        if len(content_info_records_raw) != 0x900:
            raise InvalidTMDError('Content info records length is not 0x900')

        if verify_hashes:
            real_hash = sha256(content_info_records_raw)
            if content_info_records_hash != real_hash.digest():
                raise InvalidHashError('Content Info Records hash is invalid')

        content_chunk_records_raw = fp.read(content_count * CHUNK_RECORD_SIZE)

        chunk_records = []
        for cr_raw in (content_chunk_records_raw[i:i + CHUNK_RECORD_SIZE] for i in
                       range(0, content_count * CHUNK_RECORD_SIZE, CHUNK_RECORD_SIZE)):
            chunk_records.append(ContentChunkRecord(id=cr_raw[0:4].hex(),
                                                    cindex=readbe(cr_raw[4:6]),
                                                    type=ContentTypeFlags.from_int(readbe(cr_raw[6:8])),
                                                    size=readbe(cr_raw[8:16]),
                                                    hash=cr_raw[16:48]))

        info_records = []
        for ir_raw in (content_info_records_raw[i:i + 0x24] for i in range(0, 0x900, 0x24)):
            if ir_raw != b'\0' * 0x24:
                info_records.append(ContentInfoRecord(index_offset=readbe(ir_raw[0:2]),
                                                      command_count=readbe(ir_raw[2:4]),
                                                      hash=ir_raw[4:36]))

        if verify_hashes:
            chunk_records_hashed = set()
            for ir in info_records:
                to_hash = []
                for cr in chunk_records[ir.index_offset:ir.index_offset + ir.command_count]:
                    if cr in chunk_records_hashed:
                        raise InvalidTMDError('attempting to hash chunk record twice')

                    chunk_records_hashed.add(cr)
                    to_hash.append(cr)

                hashed = sha256(b''.join(bytes(x) for x in to_hash))
                if hashed.digest() != ir.hash:
                    raise InvalidInfoRecordError(ir)

        # unused vales are loaded only for use when re-building the binary tmd
        u_issuer = header[0:0x40].decode('ascii').rstrip('\0')
        u_version = header[0x40]
        u_ca_crl_version = header[0x41]
        u_signer_crl_version = header[0x42]
        u_reserved1 = header[0x43]
        u_system_version = header[0x44:0x4C]
        u_title_type = header[0x54:0x58]
        u_group_id = header[0x58:0x5A]
        u_reserved2 = header[0x62:0x66]
        u_srl_flag = header[0x66]  # is this one used for anything?
        u_reserved3 = header[0x67:0x98]
        u_access_rights = header[0x98:0x9C]
        u_boot_count = header[0xA0:0xA2]
        u_padding = header[0xA2:0xA4]

        return cls(title_id=title_id, save_size=save_size, srl_save_size=srl_save_size, title_version=title_version,
                   info_records=info_records, chunk_records=chunk_records, signature=(sig_type, signature),
                   _u_issuer=u_issuer, _u_version=u_version, _u_ca_crl_version=u_ca_crl_version,
                   _u_signer_crl_version=u_signer_crl_version, _u_reserved1=u_reserved1,
                   _u_system_version=u_system_version, _u_title_type=u_title_type, _u_group_id=u_group_id,
                   _u_reserved2=u_reserved2, _u_srl_flag=u_srl_flag, _u_reserved3=u_reserved3,
                   _u_access_rights=u_access_rights, _u_boot_count=u_boot_count, _u_padding=u_padding)

    @classmethod
    def from_file(cls, fn: str, *, verify_hashes: bool = True) -> 'TitleMetadataReader':
        with open(fn, 'rb') as f:
            return cls.load(f, verify_hashes=verify_hashes)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

from sys import exit
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List

try:
    # noinspection PyProtectedMember
    from ._crypto import XTSN
except ImportError:
    try:
        from _crypto import XTSN
    except ImportError:
        exit("Couldn't load ccrypto. The extension needs to be compiled.")


def parse_biskeydump(keys: str):
    bis_keys: List[List[bytes]] = [[None, None], [None, None], [None, None], [None, None]]
    for l in keys.splitlines():
        # old format
        if l.startswith('BIS KEY'):
            data = l.split()[2:]
            key_idx = int(data[0])
            key_type = data[1][1:-2]
            if key_type not in {'crypt', 'tweak'}:
                raise RuntimeError('unknown type')  # TODO: give this a special error
            bis_keys[key_idx][0 if key_type == 'crypt' else 1] = bytes.fromhex(data[2])
        # new format
        elif l.startswith('bis_key'):
            keyinfo, key = l.split(" = ")
            try:
                key_idx = int(keyinfo.split("_")[2])
            except ValueError:
                continue  # some lines such as "bis_key_source_XX" may cause this exception
            bis_keys[key_idx][0] = bytes.fromhex(key[:32])
            bis_keys[key_idx][1] = bytes.fromhex(key[32:])
    return bis_keys



# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

"""
Mounts 3DSX Homebrew files, creating a virtual filesystem with the 3DSX's RomFS and SMDH.
"""

import logging
import os
from errno import ENOENT
from stat import S_IFDIR, S_IFREG
from struct import unpack
from sys import exit, argv
from typing import TYPE_CHECKING, BinaryIO

from pyctr.util import readle
from . import _common as _c
from .romfs import RomFSMount
# _common imports these from fusepy, and prints an error if it fails; this allows less duplicated code
from ._common import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context

if TYPE_CHECKING:
    from typing import Dict


class ThreeDSXMount(LoggingMixIn, Operations):
    fd = 0

    def __init__(self, threedsx_fp: BinaryIO, g_stat: os.stat_result):
        self._g_stat = g_stat
        self.g_stat = {'st_ctime': int(g_stat.st_ctime), 'st_mtime': int(g_stat.st_mtime),
                       'st_atime': int(g_stat.st_atime)}
        self.romfs_fuse: RomFSMount = None

        self.f = threedsx_fp
        threedsx_fp.seek(0, 2)
        self.total_size = threedsx_fp.tell()
        threedsx_fp.seek(0)

        header = threedsx_fp.read(0x20)
        if readle(header[4:6]) < 44:
            exit('3DSX has no SMDH or RomFS.')

        smdh_offset, smdh_size, romfs_offset = unpack('<3I', threedsx_fp.read(12))  # type: int
        self.files: Dict[str, Dict[str, int]] = {}
        if smdh_offset:  # unlikely, you can't add a romfs without this
            self.files['/icon.smdh'] = {'size': smdh_size, 'offset': smdh_offset}
        if romfs_offset:
            self.files['/romfs.bin'] = {'size': self.total_size - romfs_offset, 'offset': romfs_offset}

    def __del__(self, *args):
        try:
            self.f.close()
        except AttributeError:
            pass

    destroy = __del__

    def init(self, path):
        if '/romfs.bin' in self.files:
            try:
                romfs_vfp = _c.VirtualFileWrapper(self, '/romfs.bin', self.files['/romfs.bin']['size'])
                romfs_fuse = RomFSMount(romfs_vfp, self._g_stat)
                romfs_fuse.init(path)
                self.romfs_fuse = romfs_fuse
            except Exception as e:
                print(f'Failed to mount RomFS: {type(e).__name__}: {e}')

    def flush(self, path, fh):
        return self.f.flush()

    @_c.ensure_lower_path
    def getattr(self, path, fh=None):
        if path.startswith('/romfs/'):
            return self.romfs_fuse.getattr(_c.remove_first_dir(path), fh)
        uid, gid, pid = fuse_get_context()
        if path == '/' or path == '/romfs':
            st = {'st_mode': (S_IFDIR | 0o555), 'st_nlink': 2}
        elif path in self.files:
            st = {'st_mode': (S_IFREG | 0o444), 'st_size': self.files[path]['size'], 'st_nlink': 1}
        else:
            raise FuseOSError(ENOENT)
        return {**st, **self.g_stat, 'st_uid': uid, 'st_gid': gid}

    def open(self, path, flags):
        self.fd += 1
        return self.fd

    @_c.ensure_lower_path
    def readdir(self, path, fh):
        if path.startswith('/romfs'):
            yield from self.romfs_fuse.readdir(_c.remove_first_dir(path), fh)
        elif path == '/':
            yield from ('.', '..')
            yield from (x[1:] for x in self.files)
            if self.romfs_fuse is not None:
                yield 'romfs'

    @_c.ensure_lower_path
    def read(self, path, size, offset, fh):
        if path.startswith('/romfs/'):
            return self.romfs_fuse.read(_c.remove_first_dir(path), size, offset, fh)

        fi = self.files[path]
        real_offset = fi['offset'] + offset
        if fi['offset'] + offset > fi['offset'] + fi['size']:
            return b''
        if offset + size > fi['size']:
            size = fi['size'] - offset
        self.f.seek(real_offset)
        return self.f.read(size)

    @_c.ensure_lower_path
    def statfs(self, path):
        if path.startswith('/romfs/'):
            return self.romfs_fuse.statfs(_c.remove_first_dir(path))
        else:
            return {'f_bsize': 4096, 'f_blocks': self.total_size // 4096, 'f_bavail': 0, 'f_bfree': 0,
                    'f_files': len(self.files)}


def main(prog: str = None, args: list = None):
    from argparse import ArgumentParser
    if args is None:
        args = argv[1:]
    parser = ArgumentParser(prog=prog, description='Mount 3DSX Homebrew files.',
                            parents=(_c.default_argp, _c.main_args('threedsx', '3DSX file')))

    a = parser.parse_args(args)
    opts = dict(_c.parse_fuse_opts(a.o))

    if a.do:
        logging.basicConfig(level=logging.DEBUG, filename=a.do)

    threedsx_stat = os.stat(a.threedsx)

    with open(a.threedsx, 'rb') as f:
        mount = ThreeDSXMount(threedsx_fp=f, g_stat=threedsx_stat)
        if _c.macos or _c.windows:
            opts['fstypename'] = '3DSX'
            # assuming / is the path separator since macos. but if windows gets support for this,
            #   it will have to be done differently.
            path_to_show = os.path.realpath(a.threedsx).rsplit('/', maxsplit=2)
            if _c.macos:
                opts['volname'] = f'3DSX Homebrew ({path_to_show[-2]}/{path_to_show[-1]})'
            elif _c.windows:
                # volume label can only be up to 32 chars
                opts['volname'] = '3DSX Homebrew'
        FUSE(mount, a.mount_point, foreground=a.fg or a.do or a.d, ro=True, nothreads=True, debug=a.d,
             fsname=os.path.realpath(a.threedsx).replace(',', '_'), **opts)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

import logging
import os
from collections import defaultdict
from errno import ENOENT, EROFS
from stat import S_IFDIR, S_IFREG
from sys import argv, exit
from typing import TYPE_CHECKING
from zlib import crc32

from hac.crypto import XTSN, parse_biskeydump
from ._common import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context
from . import _common as _c

if TYPE_CHECKING:
    from typing import BinaryIO, List, TextIO

bis_key_ids = defaultdict(lambda: -1, {
    'PRODINFO': 0,
    'PRODINFOF': 0,
    'SAFE': 1,
    'SYSTEM': 2,
    'USER': 3
})


class HACNandImageMount(LoggingMixIn, Operations):
    fd = 0

    def __init__(self, nand_fp: 'BinaryIO', g_stat: os.stat_result, keys: str, readonly: bool = False):
        self.readonly = readonly
        self.g_stat = {'st_ctime': int(g_stat.st_ctime), 'st_mtime': int(g_stat.st_mtime),
                       'st_atime': int(g_stat.st_atime)}

        bis_keys = parse_biskeydump(keys)
        self.crypto: List[XTSN] = [None] * 4
        for x in range(4):
            self.crypto[x] = XTSN(*bis_keys[x])

        self.files = {}
        nand_fp.seek(0x200)
        gpt_header = nand_fp.read(0x5C)
        if gpt_header[0:8] != b'EFI PART':
            exit('GPT header magic not found.')

        header_to_hash = gpt_header[0:0x10] + b'\0\0\0\0' + gpt_header[0x14:]
        crc_expected = int.from_bytes(gpt_header[0x10:0x14], 'little')
        crc_got = crc32(header_to_hash) & 0xFFFFFFFF
        if crc_got != crc_expected:
            exit(f'GPT header crc32 mismatch (expected {crc_expected:08x}, got {crc_got:08x})')

        gpt_backup_header_location = int.from_bytes(gpt_header[0x20:0x28], 'little')
        # check if the backup header exists
        nand_fp.seek(gpt_backup_header_location * 0x200)
        gpt_backup_header = nand_fp.read(0x200)
        if gpt_backup_header[0:8] != b'EFI PART':
            exit('GPT backup header not found. This likely means an incomplete backup.')

        gpt_part_start = int.from_bytes(gpt_header[0x48:0x50], 'little')
        gpt_part_count = int.from_bytes(gpt_header[0x50:0x54], 'little')
        gpt_part_entry_size = int.from_bytes(gpt_header[0x54:0x58], 'little')

        nand_fp.seek(gpt_part_start * 0x200)
        gpt_part_full_raw = nand_fp.read(gpt_part_count * gpt_part_entry_size)
        gpt_part_crc_expected = int.from_bytes(gpt_header[0x58:0x5C], 'little')
        gpt_part_crc_got = crc32(gpt_part_full_raw) & 0xFFFFFFFF
        if gpt_part_crc_got != gpt_part_crc_expected:
            exit(f'GPT Partition table crc32 mismatch '
                 f'(expected {gpt_part_crc_expected:08x}, got {gpt_part_crc_got:08x})')
        gpt_parts_raw = [gpt_part_full_raw[i:i + gpt_part_entry_size] for i in range(0, len(gpt_part_full_raw),
                                                                                     gpt_part_entry_size)]
        for part in gpt_parts_raw:
            name = part[0x38:].decode('utf-16le').rstrip('\0')
            self.files[f'/{name.lower()}.img'] = {'real_filename': name + '.img', 'bis_key': bis_key_ids[name],
                                                  'start': int.from_bytes(part[0x20:0x28], 'little') * 0x200,
                                                  'end': (int.from_bytes(part[0x28:0x30], 'little') + 1) * 0x200}

        self.f = nand_fp

    def __del__(self, *args):
        try:
            self.f.close()
        except AttributeError:
            pass

    destroy = __del__

    def flush(self, path, fh):
        return self.f.flush()

    @_c.ensure_lower_path
    def getattr(self, path: str, fh=None):
        uid, gid, pid = fuse_get_context()
        if path == '/':
            st = {'st_mode': (S_IFDIR | (0o555 if self.readonly else 0o777)), 'st_nlink': 2}
        elif path in self.files:
            p = self.files[path]
            st = {'st_mode': (S_IFREG | (0o444 if self.readonly else 0o666)),
                  'st_size': p['end'] - p['start'], 'st_nlink': 1}
        else:
            raise FuseOSError(ENOENT)
        return {**st, **self.g_stat, 'st_uid': uid, 'st_gid': gid}

    def open(self, path: str, flags):
        self.fd += 1
        return self.fd

    @_c.ensure_lower_path
    def readdir(self, path: str, fh):
        yield from ('.', '..')
        yield from (x['real_filename'] for x in self.files.values())

    @_c.ensure_lower_path
    def read(self, path: str, size: int, offset: int, fh):
        fi = self.files[path]
        real_offset: int = fi['start'] + offset

        if fi['start'] + offset > fi['end']:
            return b''
        if offset + size > fi['end']:
            size = fi['end'] - offset

        if fi['bis_key'] >= 0:
            before = offset % 16
            after = (offset + size) % 16
            if after:
                after = 16 - after
            aligned_real_offset = real_offset - before
            aligned_offset = offset - before
            size = before + size
            self.f.seek(aligned_real_offset)
            xtsn = self.crypto[fi['bis_key']]
            return xtsn.decrypt(self.f.read(size + after), 0, 0x4000, aligned_offset)[before:size]

        else:
            self.f.seek(real_offset)
            return self.f.read(size)

    @_c.ensure_lower_path
    def write(self, path: str, data: bytes, offset: int, fh):
        if self.readonly:
            raise FuseOSError(EROFS)

        fi = self.files[path]
        real_offset: int = fi['start'] + offset
        real_len = len(data)

        if fi['start'] + offset > fi['end']:
            # not writing past the file size
            return real_len

        if real_offset + real_len > fi['end']:
            data = data[:-((real_offset + real_len) - (fi['end'] - fi['start']))]

        if fi['bis_key'] >= 0:
            before = offset % 16
            after = (offset + real_len) % 16
            aligned_offset = offset - before
            aligned_real_offset = real_offset - before
            if after:
                # this sucks...
                new_after = 16 - after
                last_block_ending = self.read(path, new_after, offset + len(data), 0)
            else:
                last_block_ending = b''

            if before:
                first_block_beginning = self.read(path, before, offset - before, 0)
            else:
                first_block_beginning = b''

            self.f.seek(aligned_real_offset)
            xtsn = self.crypto[fi['bis_key']]
            to_encrypt = b''.join((first_block_beginning, data, last_block_ending))
            self.f.write(xtsn.encrypt(to_encrypt, 0, 0x4000, aligned_offset))

        else:
            self.f.seek(real_offset)
            self.f.write(data)

        return real_len

    # TODO: get the real nand size, instead of hard-coding it
    @_c.ensure_lower_path
    def statfs(self, path: str):
        return {'f_bsize': 4096, 'f_blocks': 0x747C00000 // 4096, 'f_bavail': 0, 'f_bfree': 0,
                'f_files': len(self.files)}


def main(prog: str = None, args: list = None):
    from argparse import ArgumentParser
    if args is None:
        args = argv[1:]
    parser = ArgumentParser(prog=prog, description='Mount Nintendo Switch NAND images.',
                            parents=(_c.default_argp, _c.readonly_argp, _c.main_args('nand', 'NAND image')))
    parser.add_argument('--keys', help='keys text file from biskeydump')
    parser.add_argument('-S', '--split-files', help='treat as part of a split file', action='store_true')

    a = parser.parse_args(args)
    opts = dict(_c.parse_fuse_opts(a.o))

    if a.do:
        logging.basicConfig(level=logging.DEBUG, filename=a.do)

    def do_thing(f: 'BinaryIO', k: 'TextIO', nand_stat: os.stat_result):
        mount = HACNandImageMount(nand_fp=f, g_stat=nand_stat, keys=k.read(), readonly=a.ro)
        if _c.macos or _c.windows:
            opts['fstypename'] = 'HACFS'
            # assuming / is the path separator since macos. but if windows gets support for this,
            #   it will have to be done differently.
            if _c.macos:
                path_to_show = os.path.realpath(a.nand).rsplit('/', maxsplit=2)
                opts['volname'] = f'Nintendo Switch NAND ({path_to_show[-2]}/{path_to_show[-1]})'
            elif _c.windows:
                # volume label can only be up to 32 chars
                opts['volname'] = 'Nintendo Switch NAND'
        FUSE(mount, a.mount_point, foreground=a.fg or a.do or a.d, ro=a.ro, nothreads=True, debug=a.d,
             fsname=os.path.realpath(a.nand).replace(',', '_'), **opts)

    with open(a.keys, 'r', encoding='utf-8') as k:
        if a.split_files:
            # make sure the ending is an integer
            try:
                int(a.nand[-2:])
            except ValueError:
                exit('Could not find a part number at the end of the filename.\n'
                     'A multi-part Nintendo Switch NAND backup should have filenames in the format of '
                     '"filename.bin.XX", where XX is the part number.')

            # try to find all the parts, starting with 00
            base = a.nand[:-2]
            count = 0
            while True:
                if os.path.isfile(base + format(count, '02')):
                    count += 1
                else:
                    break

            if count == 0:
                exit('Could not find the first part of the multi-part backup.')

            handler = _c.SplitFileHandler(base + format(x, '02') for x in range(count))
            do_thing(handler, k, os.stat(base + '00'))

        else:
            with open(a.nand, 'r+b') as f:
                do_thing(f, k, os.stat(a.nand))



# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

from os.path import dirname, realpath
from sys import argv, exit, path

# path fun times
path.insert(0, dirname(realpath(__file__)))

from main import exit_print_types, mount
if len(argv) < 2:
    exit_print_types()

if argv[1] == '--install-desktop-entry':
    from main import create_desktop_entry
    prefix = None if len(argv) < 3 else argv[2]
    create_desktop_entry(prefix)
    exit(0)

exit(mount(argv.pop(1).lower()))



# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

from sys import executable
from os import remove
from os.path import abspath, dirname
from subprocess import Popen
from tempfile import NamedTemporaryFile
import winreg

# these do not use winreg due to administrator access being required.
base = r'''Windows Registry Editor Version 5.00

[HKEY_CLASSES_ROOT\*\shell\Mount with ninfs\command]
@="\"{exec}\" {extra} gui \"%1\""

[-HKEY_CLASSES_ROOT\*\shell\Mount with fuse-3ds]
'''

base_del = r'''Windows Registry Editor Version 5.00

[-HKEY_CLASSES_ROOT\*\shell\Mount with ninfs]

[-HKEY_CLASSES_ROOT\*\shell\Mount with fuse-3ds]
'''


def call_regedit(data: str):
    # this does not use winreg due to administrator access being required.
    t = NamedTemporaryFile('w', delete=False, encoding='cp1252', suffix='-ninfs.reg')
    try:
        t.write(data)
        t.close()  # need to close so regedit can open it
        p = Popen(['regedit.exe', t.name], shell=True)
        p.wait()
    finally:
        t.close()  # just in case
        remove(t.name)


def add_reg(_pyi: bool):
    fmtmap = {'exec': executable.replace('\\', '\\\\'), 'extra': ''}
    if not _pyi:
        fmtmap['extra'] = r'\"{}\"'.format(dirname(abspath(__file__)).replace('\\', '\\\\'))
    call_regedit(base.format_map(fmtmap))


def del_reg():
    call_regedit(base_del)


# I'm wondering how I could properly check if this is running as admin, instead of always checking if UAC is disabled
#   if IsUserAnAdmin is 1.
def uac_enabled() -> bool:
    key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r'SOFTWARE\Microsoft\Windows\CurrentVersion\Policies\System', 0,
                         winreg.KEY_READ)
    return bool(winreg.QueryValueEx(key, 'EnableLUA')[0])



# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.


class PyCTRError(Exception):
    """Common base class for all PyCTR errors."""



# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

"""
Mounts raw CDN contents, creating a virtual filesystem of decrypted contents (if encrypted).
"""

import logging
import os
from errno import ENOENT
from stat import S_IFDIR, S_IFREG
from sys import exit, argv
from typing import TYPE_CHECKING

from pyctr.crypto import CryptoEngine, Keyslot
from pyctr.types.tmd import TitleMetadataReader, CHUNK_RECORD_SIZE
from . import _common as _c
# _common imports these from fusepy, and prints an error if it fails; this allows less duplicated code
from ._common import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context
from .ncch import NCCHContainerMount
from .srl import SRLMount

if TYPE_CHECKING:
    from typing import Dict


class CDNContentsMount(LoggingMixIn, Operations):
    fd = 0

    # get the real path by returning self.cdn_dir + path
    def rp(self, path):
        return os.path.join(self.cdn_dir, path)

    def __init__(self, cdn_dir: str, dec_key: str = None, dev: bool = False, seeddb: str = None):
        self.cdn_dir = cdn_dir

        self.crypto = CryptoEngine(dev=dev)

        self.cdn_content_size = 0
        self.dev = dev
        self.seeddb = seeddb

        # get status change, modify, and file access times
        cdn_stat = os.stat(cdn_dir)
        self.g_stat = {'st_ctime': int(cdn_stat.st_ctime), 'st_mtime': int(cdn_stat.st_mtime),
                       'st_atime': int(cdn_stat.st_atime)}

        try:
            self.tmd = TitleMetadataReader.from_file(self.rp('tmd'))
        except FileNotFoundError:
            exit('tmd not found.')

        # noinspection PyUnboundLocalVariable
        self.title_id = self.tmd.title_id

        if not os.path.isfile(self.rp('cetk')):
            if not dec_key:
                exit('cetk not found. Provide the ticket or decrypted titlekey with --dec-key.')

        if dec_key:
            try:
                titlekey = bytes.fromhex(dec_key)
                if len(titlekey) != 16:
                    exit('--dec-key input is not 32 hex characters.')
            except ValueError:
                exit('Failed to convert --dec-key input to bytes. Non-hex character likely found, or is not '
                     '32 hex characters.')
            # noinspection PyUnboundLocalVariable
            self.crypto.set_normal_key(Keyslot.DecryptedTitlekey, titlekey)
        else:
            with open(self.rp('cetk'), 'rb') as tik:
                # load ticket
                self.crypto.load_from_ticket(tik.read(0x350))

        # create virtual files
        self.files = {'/ticket.bin': {'size': 0x350, 'type': 'raw', 'real_filepath': self.rp('cetk')},
                      '/tmd.bin': {'size': 0xB04 + self.tmd.content_count * CHUNK_RECORD_SIZE, 'offset': 0,
                                   'type': 'raw', 'real_filepath': self.rp('tmd')},
                      '/tmdchunks.bin': {'size': self.tmd.content_count * CHUNK_RECORD_SIZE, 'offset': 0xB04,
                                         'type': 'raw', 'real_filepath': self.rp('tmd')}}

        self.dirs: Dict[str, NCCHContainerMount] = {}

    def init(self, path):
        # read contents to generate virtual files
        for chunk in self.tmd.chunk_records:
            if os.path.isfile(self.rp(chunk.id)):
                real_filename = chunk.id
            elif os.path.isfile(self.rp(chunk.id.upper())):
                real_filename = chunk.id.upper()
            else:
                print(f'Content {chunk.cindex:04}:{chunk.id} not found, will not be included.')
                continue
            f_stat = os.stat(self.rp(real_filename))
            if chunk.size != f_stat.st_size:
                print('Warning: TMD Content size and filesize of', chunk.id, 'are different.')
            self.cdn_content_size += chunk.size
            is_srl = chunk.cindex == 0 and self.title_id[3:5] == '48'
            file_ext = 'nds' if is_srl else 'ncch'
            filename = f'/{chunk.cindex:04x}.{chunk.id}.{file_ext}'
            self.files[filename] = {'size': chunk.size, 'index': chunk.cindex.to_bytes(2, 'big'),
                                    'type': 'enc' if chunk.type.encrypted else 'raw',
                                    'real_filepath': self.rp(real_filename)}

            dirname = f'/{chunk.cindex:04x}.{chunk.id}'
            # noinspection PyBroadException
            try:
                content_vfp = _c.VirtualFileWrapper(self, filename, chunk.size)
                if is_srl:
                    content_fuse = SRLMount(content_vfp, g_stat=f_stat)
                else:
                    content_fuse = NCCHContainerMount(content_vfp, dev=self.dev, g_stat=f_stat, seeddb=self.seeddb)
                content_fuse.init(path)
                self.dirs[dirname] = content_fuse
            except Exception as e:
                print(f'Failed to mount {filename}: {type(e).__name__}: {e}')

    @_c.ensure_lower_path
    def getattr(self, path, fh=None):
        first_dir = _c.get_first_dir(path)
        if first_dir in self.dirs:
            return self.dirs[first_dir].getattr(_c.remove_first_dir(path), fh)
        uid, gid, pid = fuse_get_context()
        if path == '/' or path in self.dirs:
            st = {'st_mode': (S_IFDIR | 0o555), 'st_nlink': 2}
        elif path in self.files:
            st = {'st_mode': (S_IFREG | 0o444), 'st_size': self.files[path]['size'], 'st_nlink': 1}
        else:
            raise FuseOSError(ENOENT)
        return {**st, **self.g_stat, 'st_uid': uid, 'st_gid': gid}

    def open(self, path, flags):
        self.fd += 1
        return self.fd

    @_c.ensure_lower_path
    def readdir(self, path, fh):
        first_dir = _c.get_first_dir(path)
        if first_dir in self.dirs:
            yield from self.dirs[first_dir].readdir(_c.remove_first_dir(path), fh)
        else:
            yield from ('.', '..')
            yield from (x[1:] for x in self.files)
            yield from (x[1:] for x in self.dirs)

    @_c.ensure_lower_path
    def read(self, path, size, offset, fh):
        first_dir = _c.get_first_dir(path)
        if first_dir in self.dirs:
            return self.dirs[first_dir].read(_c.remove_first_dir(path), size, offset, fh)
        fi = self.files[path]
        real_size = size
        with open(fi['real_filepath'], 'rb') as f:
            if fi['type'] == 'raw':
                # if raw, just read and return
                f.seek(offset)
                data = f.read(size)

            elif fi['type'] == 'enc':
                # if encrypted, the block needs to be decrypted first
                # CBC requires a full block (0x10 in this case). and the previous
                #   block is used as the IV. so that's quite a bit to read if the
                #   application requires just a few bytes.
                # thanks Stary2001
                before = offset % 16
                if size % 16 != 0:
                    size = size + 16 - size % 16
                if offset - before == 0:
                    # use the initial value if reading from the first block
                    iv = fi['index'] + (b'\0' * 14)
                else:
                    # use the previous block if reading anywhere else
                    f.seek(offset - before - 0x10)
                    iv = f.read(0x10)
                # read to block size
                f.seek(offset - before)
                # adding 0x10 to the size fixes some kind of decryption bug
                data = self.crypto.create_cbc_cipher(Keyslot.DecryptedTitlekey,
                                                     iv).decrypt(f.read(size + 0x10))[before:real_size + before]

            else:
                from pprint import pformat
                print('--------------------------------------------------',
                      'Warning: unknown file type (this should not happen!)',
                      'Please file an issue or contact the developer with the details below.',
                      '  https://github.com/ihaveamac/ninfs/issues',
                      '--------------------------------------------------',
                      f'{path!r}: {pformat(fi)!r}', sep='\n')

                data = b'g' * size

            return data

    @_c.ensure_lower_path
    def statfs(self, path):
        first_dir = _c.get_first_dir(path)
        if first_dir in self.dirs:
            return self.dirs[first_dir].read(_c.remove_first_dir(path))
        return {'f_bsize': 4096, 'f_blocks': self.cdn_content_size // 4096, 'f_bavail': 0, 'f_bfree': 0,
                'f_files': len(self.files)}


def main(prog: str = None, args: list = None):
    from argparse import ArgumentParser
    if args is None:
        args = argv[1:]
    parser = ArgumentParser(prog=prog, description='Mount Nintendo 3DS CDN contents.',
                            parents=(_c.default_argp, _c.dev_argp, _c.seeddb_argp,
                                     _c.main_args('cdn_dir', 'directory with CDN contents')))
    parser.add_argument('--dec-key', help='decrypted titlekey')

    a = parser.parse_args(args)
    opts = dict(_c.parse_fuse_opts(a.o))

    if a.do:
        logging.basicConfig(level=logging.DEBUG, filename=a.do)

    mount = CDNContentsMount(cdn_dir=a.cdn_dir, dev=a.dev, dec_key=a.dec_key, seeddb=a.seeddb)
    if _c.macos or _c.windows:
        opts['fstypename'] = 'CDN'
        opts['volname'] = f'CDN Contents ({mount.title_id.upper()})'
    FUSE(mount, a.mount_point, foreground=a.fg or a.do or a.d, ro=True, nothreads=True, debug=a.d,
         fsname=os.path.realpath(a.cdn_dir).replace(',', '_'), **opts)



# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

"""
Mounts Executable Filesystem (ExeFS) files, creating a virtual filesystem of the ExeFS contents.
"""

import logging
import os
from errno import ENOENT
from hashlib import sha256
from stat import S_IFDIR, S_IFREG
from sys import argv
from typing import TYPE_CHECKING

from pyctr.types.exefs import ExeFSReader, ExeFSEntry, decompress_code as _decompress_code
from . import _common as _c
# _common imports these from fusepy, and prints an error if it fails; this allows less duplicated code
from ._common import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context

if TYPE_CHECKING:
    from typing import BinaryIO, Dict


class ExeFSMount(LoggingMixIn, Operations):
    fd = 0

    def __init__(self, exefs_fp: 'BinaryIO', g_stat: os.stat_result, decompress_code: bool = False, strict: bool = False):
        self.g_stat = {'st_ctime': int(g_stat.st_ctime), 'st_mtime': int(g_stat.st_mtime),
                       'st_atime': int(g_stat.st_atime)}

        self.reader = ExeFSReader.load(exefs_fp, strict)
        self.files: Dict[str, ExeFSEntry] = {'/' + x.name.replace('.', '', 1) + '.bin': x
                                             for x in self.reader.entries.values()}
        self.exefs_size = sum(x.size for x in self.reader.entries.values())
        self.code_dec = b''
        self.decompress_code = decompress_code

        self.f = exefs_fp

    def __del__(self, *args):
        try:
            self.f.close()
        except AttributeError:
            pass

    destroy = __del__

    # TODO: maybe do this in a way that allows for multiprocessing (titledir)
    def init(self, path, data=None):
        try:
            item = self.files['/code.bin']
        except KeyError:
            return  # no code, don't attempt to decompress
        else:
            if self.decompress_code:
                print('ExeFS: Decompressing .code...')
                # noinspection PyBroadException
                try:
                    self.code_dec = _decompress_code(data if data else self.read('/code.bin', item.size, item.offset, 0))
                    self.files['/code-decompressed.bin'] = ExeFSEntry(name='code-decompressed', offset=-1,
                                                                      size=len(self.code_dec),
                                                                      hash=sha256(self.code_dec).digest())
                    print('ExeFS: Done!')
                except Exception as e:
                    print(f'ExeFS: Failed to decompress .code: {type(e).__name__}: {e}')
            else:
                print('ExeFS: .code aleady decompressed')
                self.files['/code-decompressed.bin'] = ExeFSEntry(name='code-decompressed', offset=item.offset,
                                                                  size=item.size, hash=item.hash)

    @_c.ensure_lower_path
    def getattr(self, path, fh=None):
        uid, gid, pid = fuse_get_context()
        if path == '/':
            st = {'st_mode': (S_IFDIR | 0o555), 'st_nlink': 2}
        else:
            try:
                item = self.files[path]
            except KeyError:
                raise FuseOSError(ENOENT)
            st = {'st_mode': (S_IFREG | 0o444), 'st_size': item.size, 'st_nlink': 1}
        return {**st, **self.g_stat, 'st_uid': uid, 'st_gid': gid}

    def open(self, path, flags):
        self.fd += 1
        return self.fd

    @_c.ensure_lower_path
    def readdir(self, path, fh):
        yield from ('.', '..')
        yield from (x[1:] for x in self.files)

    @_c.ensure_lower_path
    def read(self, path, size, offset, fh):
        try:
            item = self.files[path]
        except KeyError:
            raise FuseOSError(ENOENT)
        if item.offset == -1:
            # special case for code-decompressed
            return self.code_dec[offset:offset + size]

        real_offset = 0x200 + item.offset + offset
        if real_offset > item.offset + item.size + 0x200:
            return b''
        if offset + size > item.size:
            size = item.size - offset
        self.f.seek(real_offset)
        return self.f.read(size)

    @_c.ensure_lower_path
    def statfs(self, path):
        return {'f_bsize': 4096, 'f_blocks': self.exefs_size // 4096, 'f_bavail': 0, 'f_bfree': 0,
                'f_files': len(self.reader)}


def main(prog: str = None, args: list = None):
    from argparse import ArgumentParser
    if args is None:
        args = argv[1:]
    parser = ArgumentParser(prog=prog, description='Mount Nintendo 3DS Executable Filesystem (ExeFS) files.',
                            parents=(_c.default_argp, _c.main_args('exefs', 'ExeFS file')))
    parser.add_argument('--decompress-code', help='decompress the .code section', action='store_true')

    a = parser.parse_args(args)
    opts = dict(_c.parse_fuse_opts(a.o))

    if a.do:
        logging.basicConfig(level=logging.DEBUG, filename=a.do)

    exefs_stat = os.stat(a.exefs)

    with open(a.exefs, 'rb') as f:
        mount = ExeFSMount(exefs_fp=f, g_stat=exefs_stat, decompress_code=a.decompress_code)
        if _c.macos or _c.windows:
            opts['fstypename'] = 'ExeFS'
            # assuming / is the path separator since macos. but if windows gets support for this,
            #   it will have to be done differently.
            path_to_show = os.path.realpath(a.exefs).rsplit('/', maxsplit=2)
            if _c.macos:
                opts['volname'] = f'Nintendo 3DS ExeFS ({path_to_show[-2]}/{path_to_show[-1]})'
            elif _c.windows:
                # volume label can only be up to 32 chars
                opts['volname'] = 'Nintendo 3DS ExeFS'
        FUSE(mount, a.mount_point, foreground=a.fg or a.do or a.d, ro=True, nothreads=True, debug=a.d,
             fsname=os.path.realpath(a.exefs).replace(',', '_'), **opts)



#!/usr/bin/env python3

# This file is a part of ninfs.
#
# Copyright (c) 2017-2019 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# You can find the full license text in LICENSE.md in the root of this project.

from importlib import import_module
from inspect import cleandoc
from os import environ, makedirs
from os.path import basename, dirname, expanduser, join as pjoin, realpath
from sys import exit, argv, path, platform, hexversion, version_info

windows = platform in {'win32', 'cygwin'}

python_cmd = 'py -3' if windows else 'python3'

mount_types = ('cci', 'cdn', 'cia', 'exefs', 'nandctr', 'nandhac', 'nandtwl', 'ncch', 'romfs', 'sd', 'srl', 'threedsx',
               'titledir')
mount_aliases = {'3ds': 'cci', '3dsx': 'threedsx', 'app': 'ncch', 'csu': 'cci', 'cxi': 'ncch', 'cfa': 'ncch',
                 'nand': 'nandctr', 'nandswitch': 'nandhac', 'nanddsi': 'nandtwl',  'nds': 'srl'}

_path = dirname(realpath(__file__))
if _path not in path:
    path.insert(0, _path)

from __init__ import __version__ as version

# this should stay as str.format so it runs on older versions
print('ninfs v{} - https://github.com/ihaveamac/ninfs'.format(version))

if hexversion < 0x030601F0:
    exit('Python {0[0]}.{0[1]}.{0[2]} is not supported. Please use Python 3.6.1 or later.'.format(version_info))


def exit_print_types():
    print('Please provide a mount type as the first argument.')
    print(' ', ', '.join(mount_types))
    print()
    print('Want to use a GUI? Use "gui" as the type! (e.g. {} -mninfs gui)'.format(python_cmd))
    exit(1)


def mount(mount_type: str, return_doc: bool = False) -> int:
    if mount_type in {'gui', 'gui_i_want_to_be_an_admin_pls'}:
        return gui(_allow_admin=mount_type == 'gui_i_want_to_be_an_admin_pls')

    # noinspection PyProtectedMember
    from pyctr.crypto import BootromNotFoundError

    if windows:
        from ctypes import windll
        if windll.shell32.IsUserAnAdmin():
            print('- Note: This should *not* be run as an administrator.',
                  '- The mount will not be normally accessible.',
                  '- This should be run from a non-administrator command prompt or PowerShell prompt.', sep='\n')
    else:
        try:
            from os import getuid
            if getuid() == 0:  # 0 == root on macos and linux
                print('- Note: This should *not* be run as root.',
                      '- The mount will not be normally accessible by other users.',
                      '- This should be run from a non-root terminal.',
                      '- If you want root to be able to access the mount,',
                      '-   you can add `-o allow_root` to the arguments.', sep='\n')
        except (AttributeError, ImportError):
            pass
    if mount_type not in mount_types and mount_type not in mount_aliases:
        exit_print_types()

    module = import_module('mount.' + mount_aliases.get(mount_type, mount_type))
    if return_doc:
        return module.__doc__

    prog = None
    if __name__ != '__main__':
        prog = 'mount_' + mount_aliases.get(mount_type, mount_type)
    try:
        return module.main(prog=prog)
    except BootromNotFoundError as e:
        print('Bootrom could not be found.',
              'Please read the README of the repository for more details.',
              'Paths checked:',
              *(' - {}'.format(x) for x in e.args[0]), sep='\n')
        return 1
    except RuntimeError as e:
        if e.args == (1,):
            pass  # assuming failed to mount and the reason would be displayed in the terminal


def create_desktop_entry(prefix: str = None):
    desktop_file = cleandoc('''
    [Desktop Entry]
    Name=ninfs
    Comment=Mount Nintendo contents
    Exec=python3 -mninfs gui
    Terminal=true
    Type=Application
    Icon=ninfs
    Categories=Utility;
    ''')
    if not prefix:
        home = expanduser('~')
        prefix = environ.get('XDG_DATA_HOME', pjoin(home, '.local', 'share'))

    app_dir = pjoin(prefix, 'applications')
    makedirs(app_dir, exist_ok=True)

    with open(pjoin(app_dir, 'ninfs.desktop'), 'w', encoding='utf-8') as o:
        print('Writing', o.name)
        o.write(desktop_file)

    for s in ('1024x1024', '128x128', '64x64', '32x32', '16x16'):
        img_dir = pjoin(prefix, 'icons', 'hicolor', s, 'apps')
        makedirs(img_dir, exist_ok=True)
        with open(pjoin(dirname(__file__), 'data', s + '.png'), 'rb') as i, \
                open(pjoin(img_dir, 'ninfs.png'), 'wb') as o:
            print('Writing', o.name)
            o.write(i.read())


def main():
    exit(mount(basename(argv[0])[6:].lower()))


def gui(_allow_admin: bool = False):
    import _gui
    return _gui.main(_allow_admin=_allow_admin)


if __name__ == '__main__':
    # path fun times

    if len(argv) < 2:
        exit_print_types()

    exit(mount(argv.pop(1).lower()))
